#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The RestoreHiddenStatesInLengthAdaptive pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np
import copy


@pattern_registry(pattern_type='RestoreHiddenStatesInLengthAdaptiveUpdateIndices')
class RestoreHiddenStatesInLengthAdaptive(Pattern):
    """The RestoreHiddenStatesInLengthAdaptive pattern.

    Fuse the original sub-graph into the custom acceleration 'RestoreHiddenStatesInLengthAdaptive' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'RestoreHiddenStatesInLengthAdaptiveUpdateIndices': [
                # minilmv2-lat-roberta
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Concat'),
                                (4, 'Reshape'), (5, 'Shape'), (6, 'ConstantOfShape'), (7, 'Mul'),
                                (8, 'Equal'), (9, 'Where'), (11, 'Expand'),
                                (12, 'ScatterElements')],
                               [(), (10, 'Unsqueeze'), (11, 'Expand')]],
                        'out': [[(0, 'Reshape'), (1, 'ExpandIndices'), (2, 'ScatterElements')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'reshape_to_3d_before_restoration',
                        1: 11,
                        2: 12,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            10: [0]
                        }], [[0], 2]],
                        2: [[{
                            12: [0],
                        }, {
                            12: [2]
                        }], [[0, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            12: [0]
                        }], [[0], 1]],
                    },
                    'returns': [10, 12]
                },
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (3, 'Concat'),
                                (4, 'Reshape'), (5, 'Equal'), (6, 'Where'), (8, 'Expand'),
                                (9, 'ScatterElements')],
                               [(), (7, 'Unsqueeze'), (8, 'Expand')]],
                        'out': [[(0, 'Reshape'), (1, 'ExpandIndices'), (2, 'ScatterElements')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'reshape_to_3d_before_restoration',
                        1: 8,
                        2: 9,
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            7: [0]
                        }], [[0], 2]],
                        2: [[{
                            9: [0],
                        }, {
                            9: [2]
                        }], [[0, 2], 3]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            9: [0]
                        }], [[0], 1]],
                    },
                    'returns': [7, 9]
                },
            ]
        }

        def _set_attr(input_indices, se_attr, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = '-1,-1,-1'
            attr['dims'] = '0,1'
            attr1 = OrderedDict()
            attr1['position'] = util.list2str(input_indices)

            reshape_3d_node_idx = model.get_node_id(node_names[0])
            model.nodes[reshape_3d_node_idx].attr = attr
            expand_indices_node_idx = model.get_node_id(node_names[1])
            model.nodes[expand_indices_node_idx].attr = attr1
            se_node_idx = model.get_node_id(node_names[2])
            model.nodes[se_node_idx].attr = se_attr

        # minilmv2-lat-roberta
        for i in range(len(pattern_mapping_config['RestoreHiddenStatesInLengthAdaptiveUpdateIndices'])):
            pattern_dict = pattern_mapping_config['RestoreHiddenStatesInLengthAdaptiveUpdateIndices'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                'RestoreHiddenStatesInLengthAdaptiveUpdateIndices', pattern_dict, model)
            if len(new_node_names) != 0:
                for i in range(len(new_node_names)):
                    attr = OrderedDict()
                    input_indices = []
                    unsqueeze_node = ret_old_nodes[i][0]
                    input_indices.append(int(unsqueeze_node.attr['axes']))
                    se_attr = ret_old_nodes[i][1].attr
                    _set_attr(input_indices, se_attr, new_node_names[i], model)
                    # the first scatter elements operation need the output of embedding layer norm
                    # but its output shape is [bsxseq_len, hidden_size]
                    # so the first scatter node need modify this tensor to 3d tensor
                    # whose shape is [bs, seq_len, hidden_size]
                    reshape_3d_node = model.get_node_by_name(new_node_names[i][0])
                    embedding_ln_out_tensor = copy.deepcopy(reshape_3d_node.output_tensors[0])
                    scatter_node = model.get_node_by_name(new_node_names[i][2])
                    # check if one input tensor is from embedding_layer_norm node
                    if scatter_node.input_tensors[0].name == reshape_3d_node.input_tensors[0].name:
                        model.change_node_input_tensors(new_node_names[i][2], 0,
                            tensor=embedding_ln_out_tensor, mode='modify')
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The AttentionBlock_AttentionMaskAddReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='AttentionBlock_AttentionMaskAddReshape')
class AttentionBlock_AttentionMaskAddReshape(Pattern):
    """The AttentionBlock_AttentionMaskAddReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionBlock_AttentionMaskAddReshape' graph.
    The search strategy is based on the following pattern mapping configs for the stable diffusionV1-5.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            # must be placed behind the QKVPreReshape pattern.
            'AttentionBlock_AttentionMaskAddReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Shape'), (1, 'Gather'), (2, 'Unsqueeze'), (12, 'Concat'),
                                (13, 'Reshape'), (14, 'Add')],
                               [(), (3, 'Shape'), (4, 'Gather'), (5, 'Unsqueeze'), (12, 'Concat')],
                               [(), (6, 'Shape'), (7, 'Gather'), (8, 'Unsqueeze'), (12, 'Concat')],
                               [(), (9, 'Shape'), (10, 'Gather'), (11, 'Unsqueeze'), (12, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Add')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 13,
                        1: 14
                    },
                    'input_tensors': {
                        0: [[{
                            13: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            14: [1]
                        }], [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[{
                            13: [0]
                        }], [[0], 1]],
                        1: [[{
                            14: [0]
                        }], [[0], 1]]
                    },
                    'returns': [12]
                },
            ]
        }

        for i in range(len(pattern_mapping_config['AttentionBlock_AttentionMaskAddReshape'])):
            pattern_dict = pattern_mapping_config['AttentionBlock_AttentionMaskAddReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping(
                "AttentionBlock_AttentionMaskAddReshape", pattern_dict, model)

            if len(new_node_names) != 0:
                logger.info('AttentionBlock_AttentionMaskAddReshape mathched...')
                logger.debug('AttentionBlock_AttentionMaskAddReshape = {}'.format(new_node_names))
                for j in range(len(new_node_names)):
                    # the first new node
                    attr = OrderedDict()
                    attr['dst_shape'] = '-1,' + '512,' + '-1,-1'
                    attr['dims'] = '0, 2, 3'
                    reshape_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[reshape_node_idx].attr = attr

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The class of objective for optimization."""

class Objective(object):
    """Objective class."""
    def __init__(self, name: str, greater_is_better: bool = True, weight_ratio: float = None):
        """Init an instance.

        Args:
            name: Objectice name.
            greater_is_better: Used to describe the usage of the metric, like: greater is better for f1, 
                this parameter is only used for quantization.
            weight_ratio: Used when there are multiple metrics, for example: you want to focus on both 
                f1 and accuracy, then you will create f1 instance and accuracy instance, and indicate 
                their weight proportion. If weight_ratio of f1 is 0.3, and weight ratio of accuracy 
                is 0.7, then the final metric to tune is f1*0.3 + accuracy*0.7, this parameter is only 
                used for quantization.
        """
        self.name = name
        self.greater_is_better = greater_is_better
        self.weight_ratio = weight_ratio

    @staticmethod
    def performance():
        """Get a performance objective."""
        return Objective(name="performance", greater_is_better=True)

    @staticmethod
    def modelsize():
        """Get a modelsize objective."""
        return Objective(name="modelsize", greater_is_better=False)


performance = Objective(name="performance", greater_is_better=True)
modelsize = Objective(name="modelsize", greater_is_better=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The neural engine operator mapping file."""

from .op import Operator, operator_registry, list2str, parseTorchListConstruct
from .tensor import Tensor
import copy

@operator_registry(operator_type='View')
class View(Operator):
    """Register the View operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

    def set_attr(self, framework, node):
        """Extract the node attr from onnxruntime."""
        if framework == 'torch':
            shape_list = []
            if node.inputsAt(1).type().kind() == 'ListType':
                shape_list = parseTorchListConstruct(node.inputsAt(1))
            else:
                for i in range(1, node.inputsSize()):
                    shape_list.append(node.inputsAt(i).toIValue())
            shape_list = [-1 if x is None else x for x in shape_list]
            self._attr['shape'] = list2str(shape_list)


#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ConvReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='ConvReshape')
class ConvReshape(Pattern):
    """The ConvReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'ConvReshape' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'ConvReshape': [
                {
                    'patterns': {
                        'in': [[(0, 'Conv'), (1, 'Shape'), (2, 'Slice'), (3, 'Concat'), (4, 'Reshape')]],
                        'out': [[(0, 'Conv'), (1, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 4
                    },
                    'input_tensors': {
                        # Conv has bias, so there are 3 inputs
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            0: [2]
                        }], [[0, 1, 2], 3]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            4: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0]
                }
            ]
        }

        def _set_attr(channel, node_names, model):
            attr = OrderedDict()
            # Dynamic batch
            # We need to reshape Conv output to (batch, channel, -1)
            attr['dst_shape'] = '-1,' + str(channel) + ',-1'
            attr['dims'] = 0
            reshape_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_node_idx].attr = attr

        for i in range(len(pattern_mapping_config['ConvReshape'])):
            pattern_dict = pattern_mapping_config['ConvReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping('ConvReshape',
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    conv_node = ret_old_nodes[j][0]
                    conv_node_idx = model.get_node_id(new_node_names[j][0])
                    model.nodes[conv_node_idx].attr = conv_node.attr
                    # Use Conv weight to get channel
                    channel = conv_node.input_tensors[1].shape[0]
                    _set_attr(channel, new_node_names[j], model)

                return model

        return model



import numpy as np
import os
import onnx
import shutil
import torch
import torch.utils.data as data
import unittest
from intel_extension_for_transformers.optimization import (
    metrics,
    objectives,
    OptimizedModel,
    QuantizationConfig,
    QuantizationMode,
    NoTrainerOptimizer,
)
from intel_extension_for_transformers.optimization.trainer import NLPTrainer
from intel_extension_for_transformers.optimization.trainer import NLPSeq2SeqTrainer
from transformers import (
    AutoModelForSequenceClassification,
    AutoTokenizer,
    AutoModelForSeq2SeqLM,
)

os.environ["WANDB_DISABLED"] = "true"
os.environ["DISABLE_MLFLOW_INTEGRATION"] = "true"
MODEL_NAME = "distilbert-base-uncased-finetuned-sst-2-english"

class DummyDataset(data.Dataset):
    def __init__(self):
        self.tokenizer = AutoTokenizer.from_pretrained(MODEL_NAME)
        self.sequence_a = "intel-extension-for-transformers is based in SH"
        self.sequence_b = "Where is intel-extension-for-transformers based? NYC or SH"
        self.encoded_dict = self.tokenizer(self.sequence_a, self.sequence_b)
        self.encoded_dict['labels'] = 1

    def __len__(self):
        return 10

    def __getitem__(self, index):
        """Returns one data pair (source and target)."""
        if index < 10:
            return self.encoded_dict


def check_onnx(model_path, dataloader):
    import onnxruntime as ort
    import numpy as np
    # Check onnxruntime
    ort_session = ort.InferenceSession(model_path)
    # Preprocess input for onnxruntime
    it = iter(dataloader)
    input = next(it)
    input_names = list(input.keys())
    for k in input_names:
        if 'label' in k:
            input.pop(k)
        else:
            input[k] = np.array(input[k])
    # Run onnxruntime inference session
    ort_session.run(None, input,)
    return True

class TestQuantization(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = AutoModelForSequenceClassification.from_pretrained(
            MODEL_NAME
        )
        self.dummy_dataset = DummyDataset()
        self.trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )
        self.optimizer = NoTrainerOptimizer(self.model)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./mlruns', ignore_errors=True)
        shutil.rmtree('./tmp_trainer', ignore_errors=True)
        shutil.rmtree('./quantized_model', ignore_errors=True)
        shutil.rmtree('fp32-model.onnx', ignore_errors=True)
        shutil.rmtree('int8-model.onnx', ignore_errors=True)

    def test_fx_model_quant(self):
        fp32_output = self.trainer.predict(self.dummy_dataset).predictions
        for mode in QuantizationMode:
            print("Quantization approach:", mode.value)
            self.trainer = NLPTrainer(
                model=self.model,
                train_dataset=self.dummy_dataset,
                eval_dataset=self.dummy_dataset,
            )

            # Check fp32 jit and onnx model, only once.
            if mode == QuantizationMode.POSTTRAININGSTATIC:
                jit_model = self.trainer.export_to_jit()
                self.trainer.export_to_onnx('fp32-model.onnx')
                self.assertTrue(check_onnx('fp32-model.onnx', self.trainer.get_eval_dataloader()))

            self.trainer.benchmark(num_of_instance=1)
            tune_metric = metrics.Metric(
                name="eval_loss", greater_is_better=False, is_relative=False, criterion=0.5
            )
            quantization_config = QuantizationConfig(
                approach=mode.name,
                metrics=[tune_metric],
                objectives=[objectives.performance]
            )
            quantized_model = self.trainer.quantize(quant_config=quantization_config, provider="inc")
            self.trainer.benchmark(self.trainer.args.output_dir, num_of_instance=1)
            # By default, model will be saved into tmp_trainer dir.
            self.trainer.save_model('./quantized_model')

            # Check int8 onnx model
            if mode == QuantizationMode.POSTTRAININGSTATIC:
                # test different configure to improve UT coverage
                self.trainer.export_to_onnx(
                    save_path=None,
                    quant_format='Qlinear',
                    dtype='S8S8',
                    opset_version=13,
                )
                self.assertTrue(check_onnx('./tmp_trainer/int8-model.onnx', self.trainer.get_eval_dataloader()))
            else:
                self.trainer.export_to_onnx('int8-model.onnx')
                self.assertTrue(check_onnx('int8-model.onnx', self.trainer.get_eval_dataloader()))

            if mode == QuantizationMode.QUANTIZATIONAWARETRAINING:
                model = onnx.load('int8-model.onnx')
                tensor_list = {tensor.name:tensor for tensor in model.graph.initializer}
                torch_data = quantized_model.classifier.state_dict()\
                                ['module._packed_params._packed_params'][0].\
                                dequantize().detach().cpu().numpy().T
                from onnx.numpy_helper import to_array
                onnx_data = to_array(tensor_list['classifier.weight_quantized'])
                onnx_scale = to_array(tensor_list['classifier.weight_scale'])
                self.assertTrue(np.allclose(torch_data, onnx_data * onnx_scale, atol=0.001))
            # Check quantized model
            output_1 = self.trainer.predict(self.dummy_dataset).predictions
            loaded_model = OptimizedModel.from_pretrained(
                './quantized_model',
            )
            self.trainer.model = loaded_model
            output_2 = self.trainer.predict(self.dummy_dataset).predictions
            self.assertTrue((fp32_output != output_1).any())

            # check loaded model
            self.assertTrue((output_1 == output_2).all())

    def test_fx_model_with_smooth_quant(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        trainer = NLPTrainer(
            model=self.model,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
        )

        tune_metric = metrics.Metric(
            name="eval_loss", greater_is_better=False, is_relative=False, criterion=0.5
        )
        quantization_config = QuantizationConfig(
            approach="PostTrainingStatic",
            metrics=[tune_metric],
            objectives=[objectives.performance],
            recipes={"smooth_quant": True,
                     "smooth_quant_args": {"alpha": 0.6},
                     }
        )
        recipes = quantization_config.recipes
        self.assertTrue(recipes["smooth_quant"])
        quantized_model = trainer.quantize(quant_config=quantization_config)
        self.assertTrue("quantize" in str(type(quantized_model.classifier.module)))
        quantization_config = QuantizationConfig(
            approach="PostTrainingStatic",
            metrics=[tune_metric],
            objectives=[objectives.performance],
            recipes={}
        )
        quantized_model = trainer.quantize(quant_config=quantization_config,
                                           train_func=train_func,
                                           eval_func=eval_func)
        self.assertTrue("quantize" in str(type(quantized_model.classifier.module)))

        with self.assertRaises(ValueError):
            quantization_config = QuantizationConfig(
                approach="PostTrainingStatic",
                metrics=[tune_metric],
                objectives=[objectives.performance],
                recipes=[]
            )

    def test_functional_quant(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        self.trainer = NLPTrainer(self.model, train_dataset=self.dummy_dataset)
        quantization_config = QuantizationConfig(
            approach='PostTrainingStatic',
            objectives=[objectives.performance]
        )
        self.trainer.quantize(quant_config=quantization_config,
                              provider="inc",
                              train_func = train_func,
                              eval_func = eval_func,)

    def test_no_trainer_quant(self):
        def eval_func(model):
            return 1

        def train_func(model):
            return model

        tune_metric = metrics.Metric(
            name="eval_loss", greater_is_better=False, is_relative=False, criterion=0.5
        )
        quantization_config = QuantizationConfig(
            approach='PostTrainingStatic',
            metrics=[tune_metric],
            objectives=[objectives.performance]
        )
        self.optimizer.eval_func = eval_func
        self.optimizer.train_func = train_func
        self.optimizer.provider = "INC"
        self.optimizer.calib_dataloader = self.trainer.get_eval_dataloader()

        opt_model = self.optimizer.quantize(quant_config=quantization_config,
                              provider="inc",
                              train_func = train_func,
                              eval_func = eval_func)

    def test_online_models(self):
        model = OptimizedModel.from_pretrained(
            'Intel/distilbert-base-uncased-finetuned-sst-2-english-int8-static'
        )

    def test_seq2seq_models(self):
        model = AutoModelForSeq2SeqLM.from_pretrained('t5-small')
        tokenizer = AutoTokenizer.from_pretrained('t5-small')
        trainer = NLPSeq2SeqTrainer(
                model=model,
                tokenizer=tokenizer,
                train_dataset=self.dummy_dataset,
                eval_dataset=self.dummy_dataset,
            )
        trainer.max_length = 128
        trainer.num_beams = 3
        trainer.metrics = metrics.Metric(
                name="eval_loss"
            )
        trainer.args.predict_with_generate = True
        trainer.builtin_eval_func(self.model)

    def test_bf16_onnx(self):
        model = AutoModelForSequenceClassification.from_pretrained(
            'google/bert_uncased_L-2_H-128_A-2'
        )
        tokenizer = AutoTokenizer.from_pretrained(
            'google/bert_uncased_L-2_H-128_A-2'
        )
        trainer = NLPSeq2SeqTrainer(
                model=model,
                tokenizer=tokenizer,
                train_dataset=self.dummy_dataset,
                eval_dataset=self.dummy_dataset,
            )
        trainer.enable_bf16 = True
        trainer.export_to_onnx()
        from onnx import TensorProto
        model = onnx.load('./tmp_trainer/bf16-model.onnx')
        for tensor in model.graph.initializer:
            if 'MatMul' in tensor.name:
                self.assertEqual(tensor.data_type, TensorProto.BFLOAT16)
                break


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Benchmark: provide the inference functions for PyTorchBenchmark and ExecutorBenchmark."""
import os
import sys
import torch
import psutil
from collections import UserDict
from .utils.utility import remove_label
from neural_compressor.utils import logger
from neural_compressor.benchmark import _Benchmark as INCBenchmark
from neural_compressor.config import BenchmarkConfig as INCBenchmarkConfig
from intel_extension_for_transformers.optimization.model import OptimizedModel


def refactor_batch_size(value, batch_size, old_batch_size=-1):
    """return batched data from value.

    Args:
        value (torch.Tensor): input data.
        batch_size (int): target batch size.
        old_batch_size (int, optional): original batch size of value. Defaults to -1.

    Returns:
        batched_value: batched data.
    """
    batched_value = value
    if isinstance(value, torch.Tensor):
        if old_batch_size == -1:
            old_batch_size = value.shape[0]
        if old_batch_size == value.shape[0]:
            if batch_size <= old_batch_size:
                batched_value = value[:batch_size]
            else:
                tmp_value = value[0].unsqueeze(0)
                for i in range(old_batch_size, batch_size):
                    batched_value = torch.cat((batched_value, tmp_value))
    return batched_value


def get_example_inputs(dataloader, batch_size=1):
    """return batched data from dataloader.

    Args:
        dataloader: dataloader.
        batch_size (int, optional): batch size. Defaults to 1.

    Returns:
        batched_data: batched data.
    """
    it = iter(dataloader)
    data = next(it)
    try:   # pragma: no cover
        for d, label in data:
            data = d
        logger.info("Label is detected in dataloader. If the detection is wrong," + \
                    "please add a fake label to avoid it.")
    except:
        pass
    old_batch_size = dataloader.batch_size
    if isinstance(data, dict) or isinstance(data, UserDict):
        batched_data = {}
        for k, v in data.items():
            batched_data[k] = refactor_batch_size(v, batch_size, old_batch_size)
    elif isinstance(data, list) or isinstance(data, tuple):   # pragma: no cover
        batched_data = []
        for v in data:
            batched_data.append(refactor_batch_size(v, batch_size, old_batch_size))
    else:   # pragma: no cover
        batched_data = refactor_batch_size(data, batch_size, old_batch_size)
    return batched_data


def preprocess_model(model, example_inputs, config, additional_cmd):
    """convert model to torchscript and generate mode.

    Args:
        model (torch.nn.Module): original model.
        example_inputs: used to jit trace original model.
        config (BenchmarkConfig): control the benchmark process.
        additional_cmd (str): additional_cmd for raw_cmd.

    Returns:
        model (torch.jit.ScriptModule): original model.
        example_inputs: preprocessed example_inputs for jit model.
        additional_cmd (str): additional_cmd for raw_cmd.
    """
    if config.generate and not config.torchscript:
        additional_cmd += " --generate"
    if config.torchscript:
        # preprocess input type to tuple or tuple of tensor
        if isinstance(example_inputs, dict) or isinstance(example_inputs, UserDict):
            example_inputs = remove_label(example_inputs)
            example_inputs = tuple(example_inputs.values())
        elif isinstance(example_inputs, list) or isinstance(example_inputs, tuple):
            example_inputs = tuple(example_inputs)

        if config.generate:
            class NewModel(torch.nn.Module):
                def __init__(self, model) -> None:
                    super().__init__()
                    self.model = model
                def forward(self, *args, **kwargs):
                    output = self.model.generate(*args, **kwargs)
                    return output
            model = NewModel(model)
            if isinstance(example_inputs, tuple):
                example_inputs = example_inputs[0] #only input_ids is used.

        with torch.no_grad():
            try:
                model = torch.jit.trace(model, example_inputs)
                model = torch.jit.freeze(model.eval())
            except:   # pragma: no cover
                model = torch.jit.trace(model, example_inputs, strict=False)
                model = torch.jit.freeze(model.eval())
    return model, example_inputs, additional_cmd


def benchmark(model_name_or_path, config=None, example_inputs=None, dataloader=None):
    """function for benchmarking model.

    Args:
        model_name_or_path (str, torch.nn.Module or torch.jit.ScriptModule): model_name_or_path.
        config (BenchmarkConfig, optional): control the benchmark process. Defaults to None.
        example_inputs (optional): used as a input for benchmarking. Defaults to None.
        dataloader (optional): used to build example_inputs for benchmarking. Defaults to None.
    """
    additional_cmd = ""
    from_pretrain_flag = False
    already_jit_flag = False

    # check model type
    if isinstance(model_name_or_path, torch.nn.Module):
        model = model_name_or_path
        if isinstance(model, torch.jit.ScriptModule):
            already_jit_flag = True
    else:
        if config.backend == 'ipex' or config.torchscript == True:
            model = OptimizedModel.from_pretrained(model_name_or_path, torchscript=True)
        else:
            model = OptimizedModel.from_pretrained(model_name_or_path)
        from_pretrain_flag = True
        if isinstance(model, torch.jit.ScriptModule):
            already_jit_flag = True
            from_pretrain_flag = False # jit model can be saved directly.

    # set kwargs to model configuration
    model.eval()
    if hasattr(model, 'config') and config.kwargs is not None:
        for k, v in config.kwargs.items():
            setattr(model.config, k, v)

    # get example inputs
    if example_inputs is None:
        assert dataloader is not None, "Please pass in an example inputs or a dataloder"
        example_inputs = get_example_inputs(dataloader, config.batch_size)
    else:
        example_inputs = refactor_batch_size(example_inputs, config.batch_size)

    # handle generate and torchscript for eager model. not for ipex
    if config.backend == 'ipex':
        additional_cmd += " --enable_ipex"
        # for ipex backend, we will convert model to torchscript if not.
        if not already_jit_flag:
            config.torchscript=True
            import intel_extension_for_pytorch as ipex
            model = ipex.optimize(model)
    if not already_jit_flag:
        model, example_inputs, additional_cmd = preprocess_model(
            model, example_inputs, config, additional_cmd
        )
        if config.torchscript:
            already_jit_flag = True

    # save preprocessed model and preprocessed input.
    tmp_model_path = os.path.join(os.getcwd() + '/tmp_model.bin')
    tmp_data_path = os.path.join(os.getcwd() + '/example_inputs.bin')
    torch.save(example_inputs, tmp_data_path)
    if already_jit_flag:
        model.save(tmp_model_path)
        additional_cmd += " --torchscript"
        # jit model can be saved directly.
        from_pretrain_flag = False
    else:
        try:
            torch.save(model, tmp_model_path)
        except Exception as e:
            assert from_pretrain_flag == True, "Please pass in " + \
              "model_name_or_path instead of torch.nn.module, due to {}".format(e)
            torch.save(model.state_dict(), tmp_model_path)
    weight_size = os.path.getsize(tmp_model_path) / 1024 / 1024
    logger.info("Model size: {} MB".format(weight_size))

    # prepare raw_command for benchmark
    if from_pretrain_flag:
        model_path = model_name_or_path
        additional_cmd += " --from_pretrain"
    else:
        model_path = tmp_model_path
    current_path = os.path.abspath(__file__).rstrip("benchmark.py")
    file_path = current_path + "utils/get_throughput.py \
        --model {} --data {} --batch_size {} --warmup {} \
        --iters {} {}".format(
        model_path, tmp_data_path, config.batch_size, config.warmup,
        config.iteration, additional_cmd
    )
    raw_cmd = "{} {}".format(sys.executable, file_path)

    # trigger INC benchmark
    if config.num_of_instance == -1:
        cpu_counts = psutil.cpu_count(logical=False)
        config.num_of_instance = int(cpu_counts // config.cores_per_instance)
    os.environ['NC_ENV_CONF'] = "False" # mark the start of benchmark
    inc_conf = INCBenchmarkConfig(
        cores_per_instance=config.cores_per_instance,
        num_of_instance=config.num_of_instance
    )
    inc_bench = INCBenchmark(inc_conf)
    inc_bench(raw_cmd=raw_cmd)

    # remove tmp files
    os.remove(tmp_model_path)
    os.remove(tmp_data_path)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
import numpy as np
import copy


@pattern_registry(pattern_type='ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements')
class ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements(Pattern):
    """The ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements pattern.

    Fuse the original sub-graph into the custom acceleration graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements': [
                # minilmv2-lat-roberta
                {
                    'patterns': {
                        'in': [[(0, 'LayerNorm'), (1, 'ExpandIndices'), (2, 'GatherElements')]],
                        'out': [[(0, 'LayerNorm'), (1, 'Reshape'), (2, 'ExpandIndices'),
                                 (3, 'GatherElements'), (4, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 0,
                        1: 'reshape_to_3d_after_layer_norm_in_attention',
                        2: 1,
                        3: 2,
                        4: 'reshape_to_2d_after_gather_elements_in_attention',
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            0: [2]
                        }], [[0, 1, 2], 3]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        2: [[{
                            1: [0]
                        }], [[0], 2]],
                        3: [[{
                            2: [0]
                        }], [[0], 2]],
                        4: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            0: [0]
                        }], [[0], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            2: [0]
                        }], [[0], 1]],
                    },
                    'returns': [0, 1, 2]
                },
            ]
        }

        def _set_attr(ln_attr, ki_attr, ge_attr, hidden_size, node_names, model):
            attr1 = OrderedDict()
            attr1['dst_shape'] = '-1,-1,' + str(hidden_size)
            attr1['dims'] = '0'
            attr2 = OrderedDict()
            attr2['dst_shape'] = '-1,' + str(hidden_size)

            ln_node_idx = model.get_node_id(node_names[0])
            model.nodes[ln_node_idx].attr = ln_attr
            reshape_3d_node_idx = model.get_node_id(node_names[1])
            model.nodes[reshape_3d_node_idx].attr = attr1
            keep_indices_node_idx = model.get_node_id(node_names[2])
            model.nodes[keep_indices_node_idx].attr = ki_attr
            gather_elements_node_idx = model.get_node_id(node_names[3])
            model.nodes[gather_elements_node_idx].attr = ge_attr
            reshape_2d_node_idx = model.get_node_id(node_names[4])
            model.nodes[reshape_2d_node_idx].attr = attr2

        # minilmv2-lat-roberta
        pattern_dict = pattern_mapping_config[
            'ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements'][0]
        model, new_node_names, ret_old_nodes = util.pattern_mapping(
            'ReshapeBeforeAndAfterAttentionOutLayerNormGatherElements', pattern_dict, model)
        if len(new_node_names) != 0:
            for i in range(len(new_node_names)):
                hidden_size = int(ret_old_nodes[i][0].input_tensors[-1].shape[0])
                ln_attr = ret_old_nodes[i][0].attr
                ki_attr = ret_old_nodes[i][1].attr
                ge_attr = ret_old_nodes[i][2].attr
                _set_attr(ln_attr, ki_attr, ge_attr, hidden_size, new_node_names[i], model)

            return model

        return model



"""pruning patterns."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from intel_extension_for_transformers.optimization.utils.utility import LazyImport
torch = LazyImport('torch')
from .logger import logger
from collections import namedtuple

PATTERNS = {}


def register_pattern(name):
    """Class decorator used to register a Pattern subclass to the registry.

    Decorator function used before a Pattern subclasses.
    Make sure that this Pattern class can be registered in PATTERNS.

    Args:
        name: A string defining the pattern type name to be used in a pruning process.

    Returns:
        cls: The class of register.
    """

    def register(pattern):
        """Register patterns."""
        PATTERNS[name] = pattern
        return pattern

    return register


def get_pattern(config, modules):
    """Get registered pattern class.

    Get a Pattern object from PATTERNS.

    Args:
        config: A config dict object that contains the pattern information.
        modules: Torch neural network modules to be pruned with the pattern.

    Returns:
        A Pattern object.

    Raises:
        AssertionError: Currently only support patterns which have been registered in PATTERNS.
    """
    name = config.pattern
    name = name.split('_')[-1]
    if "x" in name:
        return PATTERNS["NxM"](config, modules)
    if ":" in name:
        return PATTERNS["N:M"](config, modules)
    assert False, f"currently only support {PATTERNS.keys()}"  # pragma: no cover


SparsityInfo = namedtuple("SparsityInfo", ['zero_cnt', 'total_cnt', 'sparsity_ratio'])


class BasePattern:
    """Pruning Pattern.

    It defines the basic pruning unit and how this unit will be pruned during pruning, e.g. 4x1, 2:4.
    
    Args:
        config: A config dict object that contains the pattern information.
        modules: Torch neural network modules to be pruned with the pattern.

    Attributes:
        pattern: A config dict object that includes information of the pattern.    
        is_global:  A bool determining whether the pruning takes global pruning option.
                    Global pruning means that pruning scores by a pruning criterion are evaluated in all layers.
                    Local pruning, by contrast, means that pruning scores by the pruning criterion are evaluated 
                        in every layer individually.
        keep_mask_layers:A dict that includes the layers whose mask will not be updated.
        invalid_layers: The layers whose shapes don't fit the pattern.
        modules: Torch neural network modules to be pruned with the pattern.
        config: A config dict object that contains all the information including the pattern's.
        max_sparsity_ratio_per_op: A float representing the maximum sparsity that one layer could reach.
        min_sparsity_ratio_per_op: A float representing the minimum sparsity that one layer could reach.
        target_sparsity: A float representing the sparsity ratio of the modules after pruning.
    """

    def __init__(self, config, modules):
        """Initialize the basic pruning unit of a pattern."""
        self.pattern = config.pattern
        self.is_global = config.pruning_scope == "global"
        self.keep_mask_layers = {}
        self.invalid_layers = []
        self.modules = modules
        self.config = config
        self.max_sparsity_ratio_per_op = self.config['max_sparsity_ratio_per_op']
        self.min_sparsity_ratio_per_op = self.config['min_sparsity_ratio_per_op']
        self.target_sparsity_ratio = self.config['target_sparsity']
        # Not using deterministic_algorithms for all examples
        torch.use_deterministic_algorithms(False)

    def reduce_tensor(self, data, dim):
        """Reduce the data along the given dimension.
        
        Args:
            data: The input data.
            dim: The reduced axis.

        Returns:
            The reduced tensor.
        """
        name = self.config['criterion_reduce_type']
        if name == "mean":
            return torch.mean(data, dim=dim)
        elif name == "sum":  # pragma: no cover
            return torch.sum(data, dim=dim)
        elif name == "max":  # pragma: no cover
            return torch.max(data, dim=dim)[0]
        else:  # pragma: no cover
            assert False, "currently only support mean, sum and max reduce type"

    def get_masks(self, scores, target_sparsity_ratio, pre_masks):
        """Generate the weight masks according to the weight score and the current target sparsity ratio.

        Args:
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            target_sparsity_ratio: A float representing the sparsity of the modules after pruning.
            pre_masks: A dict{"layer_name": Tensor} that stores the masks generated at last pruning step.

        Returns:
            A dict with the identical size as pre_masks and its 0/1 values are updated. 
                1 means unpruned and 0 means pruned.
        """
        if self.is_global:
            return self.get_masks_global(scores, target_sparsity_ratio, pre_masks)
        else:
            return self.get_masks_local(scores, target_sparsity_ratio, pre_masks)

    def get_masks_global(self, scores, target_sparsity_ratio, pre_masks):
        """Generate the weight masks for global pruning, please refer to function get_masks for more information."""
        raise NotImplementedError

    def get_masks_local(self, scores, target_sparsity_ratio, pre_masks):
        """Generate the weight masks for local pruning.
        
        Args:
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            target_sparsity_ratio: A float. After pruning, the sparsity of the modules will reach this value.
            pre_masks: A dict{"layer_name": Tensor}. The previous masks generated at the last pruning step.

        Returns:
            A dict with the identical size as pre_masks and its 0/1 values are updated. 
                1 means unpruned and 0 means pruned.
        """
        masks = {}
        if isinstance(self, PatternNxM) and not isinstance(self.block_size, dict):  # pragma: no cover
            self.block_size = self.get_block_size_dict(pre_masks)
        for key in scores.keys():
            score = {key: scores[key]}
            pre_mask = {key: pre_masks[key]}
            mask = self.get_masks_global(score, target_sparsity_ratio, pre_mask)
            masks[key] = mask[key]
        return masks

    def get_single_mask_per_target_ratio(self, score, exact_sparsity_ratio):
        """Generate a mask for one layer with the exact_sparsity_ratio.

        Args:
            score: A Tensor representing the pruning scores of each weight elements.
            exact_sparsity_ratio: A float representing the layer's final sparsity ratio.

        Returns:
            A Tensor with the identical size as score. a new mask.
        """
        flattern_score = torch.flatten(score)
        k = int(exact_sparsity_ratio * flattern_score.numel())
        threshold, _ = torch.kthvalue(flattern_score, k)
        if not k < 1:
            zero = torch.tensor([0.]).to(score.device)
            one = torch.tensor([1.]).to(score.device)
            mask = torch.where(score <= threshold, zero, one)
        else:  # pragma: no cover
            mask = torch.ones(score.shape, device=score.device)
        return mask

    def get_block_size_dict(self, data):
        """Get pattern size for each module.
        
        This is mainly for per-channel pruning when each module has different pruning size.
        
        Args:
            data: the input data.

        Returns:
            To be implemented in subclasses.
        """
        raise NotImplementedError

    def get_sparsity_ratio(self, pre_masks, return_dict=False):
        """Calculate the zero elements' ratio in pre_masks.

        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            return_dict: A bool determining whether to return more information like zero_cnt and total_cnt.

        Returns:
            A float representing the zero elements' ratio in pre_masks.
        """
        zero_cnt = 0 # This function might need to refactor in subclass.
        total_cnt = 0
        for key in pre_masks.keys():
            pre_mask = pre_masks[key]
            zero_cnt += torch.sum(pre_mask == 0.0).data.item()
            total_cnt += pre_masks[key].numel()  ##FIXME
        if return_dict:  # pragma: no cover
            return {"sparsity_ratio": float(zero_cnt) / total_cnt, "zero_cnt": zero_cnt, "total_cnt": total_cnt}
        else:
            return float(zero_cnt) / total_cnt

    def get_pattern_lock_masks(self, modules):
        """Obtain masks from original weight map according the pattern and weights' zero positions.

        Args:
            modules: a dict {'layer_name': Tensor} that stores weights.

        Returns:
            A dict with the identical size as modules, containing pattern lock masks.
        """
        pattern_lock_masks = {}
        for key in modules.keys():
            weight = modules[key].weight
            shape = weight.shape
            mask = torch.ones(shape)
            mask[weight == 0] = 0.0
            pattern_lock_masks[key] = mask.to(weight.device)
        return pattern_lock_masks

    def check_layer_validity(self):
        """Check if a layer is valid for this block_size."""
        pass

    def get_reduced_masks_from_data(self, data, key):
        """Obtain the unpruned weights and reshape according to the block_size."""
        raise NotImplementedError

    def update_residual_cnt(self, masks, target_sparsity_ratio):
        """Update the number of parameters yet to be pruned.
        
        Args:
            masks: the current pruning mask.
            target_sparsity_ratio: A float representing the final sparsity of the modules.

        Returns:
            An int representing the number of weights left to be pruned to reach the target sparsity ratio.
        """
        self.total_params_cnt = self.get_sparsity_ratio(masks, return_dict=True)["total_cnt"]
        to_prune_cnt = int(self.total_params_cnt * target_sparsity_ratio)
        for key in masks.keys():
            if self.keep_mask_layers.get(key, False):  # pragma: no cover
                zero_cnt = self.get_sparsity_ratio({key: masks[key]}, return_dict=True)["zero_cnt"]
                to_prune_cnt -= zero_cnt

        return to_prune_cnt

    def get_sparsity_ratio_each_layer(self, masks):
        """Calculate the sparsity ratio of each layer.
        
        Args:
            masks: The current weight masks.

        Returns:
            infos: the sparsity information for each layer including sparsity_ratio, zero_point and total cnts.
            SparsityInfo: the sparsity information for the model.
        """
        infos = {}
        zero_cnts = 0
        total_cnts = 0
        for key in masks.keys():
            if key in self.invalid_layers:  # pragma: no cover
                continue
            reduced_mask = self.get_reduced_masks_from_data(masks[key], key)
            zero_cnt = (int(torch.sum(reduced_mask == 0.0).data.item()))
            total_cnt = int(reduced_mask.numel())
            sparsity_ratio = float(zero_cnt) / total_cnt
            val = SparsityInfo(zero_cnt, total_cnt, sparsity_ratio)
            infos[key] = val
            zero_cnts += zero_cnt
            total_cnts += total_cnt
        sparsity_ratio = float(zero_cnts) / total_cnts
        return infos, SparsityInfo(zero_cnts, total_cnts, sparsity_ratio)

    def adjust_ratio(self, masks: dict, layer_name: str, key_new_sparsity: SparsityInfo,
                     max_sparsity_ratio: float, min_sparsity_ratio: float, \
                     final_target_sparsity_ratio: float):
        """Adjust the sparsity of a layer based on threshold.
        
        Args:
            masks: The weight masks.
            layer_name: The layer to be examined.
            key_new_sparsity: The proposed ratio for the layer.
            max_sparsity_ratio: A float representing the maximum sparsity that one layer could reach.
            min_sparsity_ratio: A float representing the minimum sparsity that one layer could reach.
            final_target_sparsity_ratio: The final target sparsity ratio.

        Returns:
            A bool indicating if the ratio needs to be adjusted.
            adjust_sparsity_ratio: The adjusted sparsity ratio.
        """
        need_adjust = False
        adjust_zero_cnt = key_new_sparsity.zero_cnt
        adjust_sparsity_ratio = key_new_sparsity.sparsity_ratio
        adjust_total_cnt = key_new_sparsity.total_cnt

        if adjust_sparsity_ratio > max_sparsity_ratio:  # pragma: no cover
            need_adjust = True
            adjust_sparsity_ratio = max_sparsity_ratio
            adjust_zero_cnt = int(adjust_total_cnt * max_sparsity_ratio)

        if adjust_sparsity_ratio < min_sparsity_ratio:  # pragma: no cover
            return need_adjust, adjust_sparsity_ratio

        ##TODO no need to calculate each time
        infos, net_info = self.get_sparsity_ratio_each_layer(masks)

        any_exceed_target_ratio = False
        for key in infos.keys():
            if infos[key].sparsity_ratio > final_target_sparsity_ratio:
                any_exceed_target_ratio = True
                break
        if adjust_sparsity_ratio > final_target_sparsity_ratio:
            any_exceed_target_ratio = True
        if not any_exceed_target_ratio:
            return need_adjust, adjust_sparsity_ratio

        zero_cnt_below_min_sparsity = 0
        total_cnt_below_min_sparsity = 0
        zero_cnt_above_min_sparsity = 0
        for key in infos.keys():
            info = infos[key]
            if key == layer_name:
                info = SparsityInfo(zero_cnt=adjust_zero_cnt, total_cnt=adjust_total_cnt,
                                    sparsity_ratio=adjust_sparsity_ratio)
            if info.sparsity_ratio < min_sparsity_ratio:
                zero_cnt_below_min_sparsity += info.zero_cnt
                total_cnt_below_min_sparsity += info.total_cnt
            else:
                zero_cnt_above_min_sparsity += info.zero_cnt

        gap_cnt = int(total_cnt_below_min_sparsity * min_sparsity_ratio) - zero_cnt_below_min_sparsity
        remaining_cnt = int(net_info.total_cnt * final_target_sparsity_ratio) \
                        - zero_cnt_above_min_sparsity - zero_cnt_below_min_sparsity
        if remaining_cnt >= gap_cnt:
            return need_adjust, adjust_sparsity_ratio
        else:  # pragma: no cover
            new_zero_cnt = adjust_zero_cnt - (gap_cnt - remaining_cnt)
            new_sparsity_ratio = float(new_zero_cnt) / adjust_total_cnt
            ##adjust_zero_cnt = new_zero_cnt
            adjust_sparsity_ratio = new_sparsity_ratio
            return True, adjust_sparsity_ratio


@register_pattern('NxM')
class PatternNxM(BasePattern):
    """Pruning Pattern.
    
    A Pattern class derived from BasePattern. In this pattern, the weights in a NxM block will be pruned or kept
    during one pruning step.
    
    Args:
        config: A config dict object that contains the pattern information.
        
    Attributes:
            block_size: A list of two integers representing the height and width of the block.
            Please note that the vertical direction of a Linear layer's weight refers to the output channel.
                because PyTorch's tensor matmul has a hidden transpose operation.
    """

    def __init__(self, config, modules):
        """Initialize the basic pruning unit of NXM pattern."""
        super(PatternNxM, self).__init__(config, modules)
        pattern = self.pattern.split('_')[-1]
        self.N = pattern.split('x')[0]
        self.M = pattern.split('x')[1]
        if self.N == "channel":  # pragma: no cover
            self.block_size = ["channel", int(self.M)]
        elif self.M == "channel":  ##channel-wise pruning mode
            self.block_size = [int(self.N), "channel"]
        else:
            self.block_size = [int(pattern.split('x')[0]), int(pattern.split('x')[1])]
        self.total_params_cnt = -1

        self.block_size = self.get_block_size_dict()
        self.check_layer_validity()

    def get_block_size_dict(self):
        """Calulate the zero elements' ration in pre_masks.
        
        Args:
            data: Dict{"layer_name": Tensor} that stores weights or scores.
            
        Returns:
            A dict. Dict{"layer_name": [block_size_1, block_size_2]} containing block shapes of each layer.
                In channel-wise pruning different layers can have different pruning patterns.
        """
        data = self.modules
        block_sizes_dict = {}
        if self.N == "channel" or self.M == "channel":
            for key in data.keys():
                if isinstance(data[key], torch.nn.Module):
                    shape = data[key].weight.shape
                else:
                    shape = data[key].shape
                if self.N == "channel":
                    block_sizes_dict[key] = [shape[0], 1]
                else:
                    block_sizes_dict[key] = [1, shape[1]]
            return block_sizes_dict
        for key in data.keys():
            block_sizes_dict[key] = self.block_size
        return block_sizes_dict

    def check_layer_validity(self):
        """Check if a layer is valid for this block_size."""
        block_sizes = self.block_size
        datas = self.modules
        for key in datas.keys():
            data = datas[key].weight
            data = self._reshape_orig_to_2dims(data)
            shape = data.shape
            block_size = block_sizes[key]
            if shape[0] % block_size[0] != 0 or shape[1] % block_size[1] != 0:  ## only consider input channel
                self.invalid_layers.append(key)
                logger.warning(f"{key} shape {data.shape} cannot be divided by {self.pattern}")

    def get_reduced_masks_from_data(self, data, key):
        """Obtain the unpruned weights and reshape according to the block_size.
        
        Args:
            data: Input.
            key: The layer name.
        
        Returns:
            The unpruned weights.
        """
        assert key not in self.invalid_layers
        block_size = self.block_size[key]
        data = self._reshape_orig_to_2dims(data)
        shape = data.shape
        new_shape = [shape[0] // block_size[0], block_size[0], shape[1] // block_size[1], block_size[1]]
        data = data.reshape(new_shape)
        data = data.sum(-1).sum(1)
        reduced_mask = data != 0
        return reduced_mask

    def get_sparsity_ratio(self, pre_masks, return_dict=False):
        """Please note that the zero cnt and total cnt are all block_wise for supporting channel-wise pruning.

        Args:
            pre_masks: Dict{"layer_name": Tensor} representing the masks generated after the last pruning step.
            return_dict: A bool determining whether to return more information like zero_cnt and total_cnt.

        Returns:
            A float representing the zero elements' ratio in pre_masks.
        """
        zero_cnt = 0
        total_cnt = 0
        for key in pre_masks.keys():
            if key in self.invalid_layers:
                continue
            reduced_mask = self.get_reduced_masks_from_data(pre_masks[key], key)
            zero_cnt += (int(torch.sum(reduced_mask == 0.0).data.item()))
            total_cnt += int(reduced_mask.numel())
        if total_cnt == 0:
            sparsity_ratio = 0.0
        else:
            sparsity_ratio = float(zero_cnt) / total_cnt
        if return_dict:
            return {"sparsity_ratio": sparsity_ratio, "zero_cnt": zero_cnt, "total_cnt": total_cnt}
        else:
            return sparsity_ratio

    def get_sparsity_ratio_progressive(self, pre_masks, return_dict=False):
        """Calculate the sparsity ratio of each layer.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            return_dict: A bool determining whether to return more information like zero_cnt and total_cnt.
        
        Returns:
            A float representing the zero elements' ratio in pre_masks.
        """
        zero_cnt = 0
        total_cnt = 0
        for key in pre_masks.keys():
            if key in self.invalid_layers:
                continue
            # progressive masks are unstructured, therefore directly find zeros
            zero_cnt += float(torch.sum(pre_masks[key] == 0).data.item())
            total_cnt += float(pre_masks[key].numel())
        return (zero_cnt / total_cnt)

    def _reshape_orig_to_2dims(self, data):
        """Process layers that are not two-dimensional(e.g conv layer).

        Args:
            data: Input.
            
        Returns:
            Reshaped data.
        """
        ##TODO need to verify whether it's ok for transposed conv
        if len(data.shape) == 4:
            data = data.permute(0, 2, 3, 1)  ##cout,k,k,cin
            data = data.reshape(data.shape[0], -1)
        return data

    def _reshape_2dims_to_orig(self, data, orig_shape):
        """Recover layers that are not two-dimensional(e.g conv layer).

        Args:
            data: Input.
            orig_shape: Target shape.
            
        Returns:
            Reshaped data.
        """
        if len(orig_shape) == 4:
            data = data.reshape(orig_shape[0], orig_shape[2], orig_shape[3],
                                orig_shape[1])
            data = data.permute(0, 3, 1, 2)
        return data

    def reshape_orig_to_pattern(self, data, key):
        """Reshape the data(s1,s2) to [s1/N,N,s2,s2/M].

        Args:
            data: The input.
            key: The layer name.
            
        Returns:
            Reshaped input tensor.
        """
        block_size = self.block_size[key]
        data = self._reshape_orig_to_2dims(data)
        shape = data.shape
        new_shape = [shape[0] // block_size[0], block_size[0], shape[1] // block_size[1],
                     block_size[1]]
        data = data.reshape(new_shape)
        return data

    def reshape_reduced_to_orig(self, data, key, orig_shape):
        """Reshape the data [s1/N,s2/M] to [s1,s2], also permute dims for conv layer.

        Args:
            data: Input.
            key: The layer name.
            orig_shape: The original shape of the layer.
            
        Returns:
            Data of its original shape.
        """
        block_size = self.block_size[key]
        data = data.repeat_interleave(block_size[0], dim=0).repeat_interleave(block_size[1], dim=-1)
        data = self._reshape_2dims_to_orig(data, orig_shape)
        return data

    def reduce_scores(self, scores):
        """Recalculate the pruning scores after reducing the data.
        
        Args:
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.

        Returns:
            The reduced pruning scores.
        """
        new_scores = {}
        for key in scores.keys():
            if key in self.invalid_layers:  # pragma: no cover
                continue
            if self.keep_mask_layers.get(key, False):
                continue
            self.keep_mask_layers[key] = False
            current_score = scores[key]
            current_score = self.reshape_orig_to_pattern(current_score, key)
            ##sum or mean is quite different for per channel pruning
            current_score_sum = self.reduce_tensor(self.reduce_tensor(current_score, dim=-1), dim=1)
            new_scores[key] = current_score_sum
        return new_scores

    def get_mask_per_threshold(self, score, threshold, block_size):
        """Get the mask per threshold."""
        zero = torch.tensor([0.]).to(score.device)
        one = torch.tensor([1.]).to(score.device)
        mask = torch.where(score <= threshold, zero, one)
        mask = mask.repeat_interleave(block_size[0], dim=0).repeat_interleave(block_size[1], dim=-1)
        return mask

    def get_masks_global(self, scores, cur_target_sparsity_ratio, pre_masks,
                         keep_exact_sparsity_ratio=True):
        """Generate masks for layers.

        Gather all layer's scores together and calculate a common threshold.
        This threshold will be applied to all layers.
        
        Args:
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            cur_target_sparsity_ratio: A float representing the model's sparsity after pruning.
            pre_masks: A dict{"layer_name": Tensor} that stores the masks generated at the last pruning step.
            max_sparsity_ratio_per_op: A float representing the maximum sparsity that one layer can reach.
            keep_pre_masks: A bool representing if the masks should remain unchanged.
            
        Returns:
            A dict with the identical size as pre_masks and its 0/1 values are updated. 
                1 means unpruned and 0 means pruned.
        """
        ##keep the masks if the layer exceed max sparsity ratio

        masks = pre_masks

        k_blockwise = self.update_residual_cnt(masks, cur_target_sparsity_ratio)
        if k_blockwise <= 0:
            return masks
        new_scores = self.reduce_scores(scores)
        global_scores = torch.cat([torch.flatten(v) for v in new_scores.values()])
        residual_k = k_blockwise
        not_exceed_layers = [key for key in new_scores.keys()]
        if self.min_sparsity_ratio_per_op > 0:
            sparsity_infos_perlayer, _ = self.get_sparsity_ratio_each_layer(masks)

        while True:
            threshold, _ = torch.kthvalue(global_scores, residual_k)
            for key in not_exceed_layers:
                block_size = self.block_size[key]
                score = new_scores[key]
                mask = self.get_mask_per_threshold(score, threshold, block_size)
                info = self.get_sparsity_ratio({key: mask}, return_dict=True)
                zero_cnt = info["zero_cnt"]
                total_cnt = info["total_cnt"]
                current_sparsity_ratio = float(zero_cnt) / total_cnt
                key_new_sparsity = SparsityInfo(zero_cnt, total_cnt, current_sparsity_ratio)
                need_adjust, adjust_ratio = self.adjust_ratio(masks, key, key_new_sparsity,
                                                              self.max_sparsity_ratio_per_op,
                                                              self.min_sparsity_ratio_per_op,
                                                              self.target_sparsity_ratio)
                if need_adjust:  # pragma: no cover
                    # uptade status
                    self.keep_mask_layers[key] = True
                    masks[key] = self.get_single_mask_per_target_ratio(new_scores[key], adjust_ratio)
                    masks[key] = masks[key].repeat_interleave(block_size[0], 0).repeat_interleave(block_size[1], -1)
                    if keep_exact_sparsity_ratio:
                        zero_cnt = self.get_sparsity_ratio({key: masks[key]}, return_dict=True)["zero_cnt"]
                        residual_k -= zero_cnt
                else:
                    masks[key] = mask
            if not keep_exact_sparsity_ratio:  # pragma: no cover
                break
            new_not_exceed_layers = [key for key in new_scores.keys() if not self.keep_mask_layers.get(key, False)]
            if not_exceed_layers == new_not_exceed_layers or len(new_not_exceed_layers) == 0:  # pragma: no cover
                break
            not_exceed_layers = new_not_exceed_layers
            global_scores = torch.cat([torch.flatten(new_scores[key]) for key in not_exceed_layers])

        for key in masks.keys():
            if key in self.invalid_layers:  # pragma: no cover
                continue
            if len(scores[key].shape) == 4:  ## need to permute
                mask = masks[key]
                orig_shape = scores[key].shape
                mask = self._reshape_2dims_to_orig(mask, orig_shape)
                masks[key] = mask
            layer_ratio = torch.sum(masks[key] == 0.0).data.item() / masks[key].numel()
            logger.info(f'layer {key} sparsity_ratio is {layer_ratio}')
        return masks

    def get_pattern_lock_masks(self, modules):
        """Obtain masks from original weight map by masking the zero-valued weights.
        
        Args:
            modules: A dict{"layer_name": Tensor} that stores weights.
            
        Returns:
            A dict with the identical size as modules, containing pattern lock masks.
        """
        pattern_lock_masks = {}
        for key in modules.keys():
            weight = modules[key].weight
            ori_shape = weight.shape
            if key in self.invalid_layers:
                mask = torch.ones(weight.shape, device=weight.device)
                pattern_lock_masks[key] = mask
                continue
            reduced_mask = self.get_reduced_masks_from_data(weight, key)
            mask = self.reshape_reduced_to_orig(reduced_mask, key, ori_shape)
            pattern_lock_masks[key] = mask
        return pattern_lock_masks

    # ---------------progressive related--------------------
    def count_new_masked_cnts(self, new_added_masks):
        """Count the number of elements to be masked.
        
        Args:
            new_added_masks: A dict {"layer_name": Tensor} that stores the added masks.

        Returns:
            The number of masked weights.
        """
        # count how many elements are to masked,
        new_masked_cnts = 0
        for key in new_added_masks.keys():
            new_masked_cnts += torch.nonzero(1 - new_added_masks[key]).size()[0]
        return new_masked_cnts

    def update_new_added_masks(self, pre_masks, cur_masks):
        """Obtain the new set-to-zero masks during a pruning procedure.

        Pre_masks, cur_masks should have identical keys bacause they represent the same model.

        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
        
        Returns:
            A dict {"layer_name": Tensor} that stores the added masks.
        """
        # obtain the new set-to-zero mask during a pruning procedure.
        # pre_masks, cur_masks should have identical keys bacause they stands for one model.
        new_added_masks = {}
        for key in pre_masks.keys():
            pre_mask = pre_masks[key]
            cur_mask = cur_masks[key]
            zero = torch.tensor([0.]).to(pre_mask.device)
            one = torch.tensor([1.]).to(cur_mask.device)
            new_added_masks[key] = torch.where(pre_mask == cur_mask, one, zero)
        return new_added_masks

    def update_progressive_masks(self, pre_masks, cur_masks, scores, progressive_step, progressive_configs):
        """Generate the progressive masks.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            progressive_step: An integer representing the number of current step in progressive pruning.
            progressive_configs: A dict that stores configurations of progressive pruning.
        
        Returns:
            A dict{"layer_name": Tensor} that stores the masks generated in progressive pruning.
        """
        use_global = progressive_configs["use_global"]
        if use_global:
            return self.update_progressive_masks_global(pre_masks, cur_masks, scores, \
                                                        progressive_step, progressive_configs)
        else:  # pragma: no cover
            return self.update_progressive_masks_local(pre_masks, cur_masks, scores, \
                                                       progressive_step, progressive_configs)

    def update_progressive_masks_linear(self, pre_masks, cur_masks, 
           progressive_step, progressive_configs):  # pragma: no cover
        """Generate the progressive masks along the block's larger dimension.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
            progressive_step: An integer representing the number of current step in progressive pruning.
            progressive_configs: A dict that stores configurations of progressive pruning.

        Returns:
            A dict{"layer_name": Tensor} that stores the masks generated in progressive pruning.
        """
        progressive_steps = progressive_configs["progressive_steps"]
        progressive_masks = {}
        new_added_masks = self.update_new_added_masks(pre_masks, cur_masks)
        for key in pre_masks.keys():
            block_size = self.block_size[key]
            new_added_mask = new_added_masks[key]
            # conv
            new_added_mask = self._reshape_orig_to_2dims(new_added_mask)
            shape = new_added_mask.shape
            # progressive masks are generated in the direction of block's large dim.
            if block_size[0] >= block_size[1]:
                # NxM (N>=M), output channel pruning
                new_shape = [shape[0] // block_size[0], progressive_steps, block_size[0] // progressive_steps,
                             shape[1] // block_size[1], block_size[1]]
                new_added_mask_reshape = new_added_mask.reshape(new_shape)
                new_added_mask_reshape[:, progressive_step:, :, :, :] = 1.0
            else:
                # NxM (N<M), input channel pruning
                new_shape = [shape[0] // block_size[0], block_size[0], shape[1] // block_size[1],
                             progressive_steps, block_size[1] // progressive_steps]
                new_added_mask_reshape = new_added_mask.reshape(new_shape)
                new_added_mask_reshape[:, :, :, progressive_step:, :] = 1.0
            new_added_mask = new_added_mask_reshape.reshape(shape)
            new_added_mask = self._reshape_2dims_to_orig(new_added_mask, pre_masks[key].shape)
            progressive_masks[key] = pre_masks[key] * new_added_mask
        return progressive_masks

    def update_progressive_masks_scores(self, pre_masks, cur_masks, scores, progressive_step, progressive_configs):
        """Generate the progressive masks based on scores.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            progressive_step: An integer representing the number of current step in progressive pruning.
            progressive_configs: A dict that stores configurations of progressive pruning.

        Returns:
            A dict{"layer_name": Tensor} that stores the masks generated in progressive pruning.
        """
        progressive_steps = progressive_configs["progressive_steps"]
        progressive_masks = {}
        new_added_masks = self.update_new_added_masks(pre_masks, cur_masks)
        for key in scores.keys():
            block_size = self.block_size[key]
            mask_num_each_block = progressive_step * int((block_size[0] * block_size[1]) / progressive_steps)
            new_added_filter = 1 - new_added_masks[key]
            score = scores[key]
            score_masked = (score * new_added_filter).abs()
            score_masked = self._reshape_orig_to_2dims(score_masked)

            # similar to n:m type
            # generate progressive masks from scores
            shape = score_masked.shape
            new_shape = [shape[0] // block_size[0], block_size[0], shape[1] // block_size[1], block_size[1]]
            score_masked_new = score_masked.clone()
            score_masked_new_shape = score_masked_new.reshape(new_shape)
            score_masked_new_shape = score_masked_new_shape.permute(0, 2, 1, 3)
            score_masked_new_flatten = torch.flatten(score_masked_new_shape, start_dim=2, end_dim=3)
            threshold, _ = torch.kthvalue(score_masked_new_flatten, mask_num_each_block, dim=2)
            threshold = threshold.unsqueeze(-1)
            threshold = threshold.repeat(1, 1, (block_size[0] * block_size[1]))
            threshold = threshold.reshape([threshold.shape[0], threshold.shape[1], block_size[0], block_size[1]])
            threshold = threshold.permute(0, 2, 1, 3)
            threshold = threshold.reshape((shape[0], shape[1]))
            one = torch.tensor([1.]).to(score.device)
            zero = torch.tensor([0.]).to(score.device)
            mask = torch.where(score_masked <= threshold, zero, one)

            mask = self._reshape_2dims_to_orig(mask, pre_masks[key].shape)
            progressive_mask = pre_masks[key] * (mask + new_added_masks[key])
            progressive_masks[key] = torch.where(progressive_mask == 0, zero, one)  # binary
        return progressive_masks

    def update_progressive_masks_local(self, pre_masks, cur_masks, scores, progressive_step, progressive_configs):
        """Generate progressive masks in a local pruning domain.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            progressive_step: An integer representing the number of current step in progressive pruning.
            progressive_configs: A dict that stores configurations of progressive pruning.
        
        Returns:
            A dict{"layer_name": Tensor} that stores the masks generated in progressive pruning.
        """
        progressive_type = progressive_configs["progressive_type"]
        if progressive_type == "linear":
            progressive_masks = self.update_progressive_masks_linear(pre_masks, cur_masks, \
                                                                     progressive_step, progressive_configs)
        elif progressive_type == "scores":
            progressive_masks = self.update_progressive_masks_scores(pre_masks, cur_masks, scores, \
                                                                     progressive_step, progressive_configs)
        else:
            raise NotImplementedError
        return progressive_masks

    def update_progressive_masks_global(self, pre_masks, cur_masks, scores, progressive_step, progressive_configs):
        """Gather all layer's scores to obtain a threshold that would be applied to all layers.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            cur_masks: Dict{"layer_name": Tensor} that stores the current masks.
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            progressive_step: An integer representing the number of current step in progressive pruning.
            progressive_configs: A dict that stores configurations of progressive pruning.
        
        Returns:
            A dict{"layer_name": Tensor} that stores the masks generated in progressive pruning.
        """
        progressive_steps = progressive_configs["progressive_steps"]
        progressive_masks = {}
        global_new_added_score_list = []
        new_added_masks = self.update_new_added_masks(pre_masks, cur_masks)
        new_added_masks_cnts = self.count_new_masked_cnts(new_added_masks)
        kth_masked_position = (new_added_masks_cnts * progressive_step) // progressive_steps
        for key in scores.keys():
            block_size = self.block_size[key]
            mask_num_each_block = progressive_step * int((block_size[0] * block_size[1]) // progressive_steps)
            new_added_filter = 1 - new_added_masks[key]
            new_added_cnts = torch.nonzero(new_added_filter).size()[0]
            score = scores[key]

            score_masked = (score * new_added_filter).abs()
            score_masked_row, _ = torch.sort(score_masked.flatten(), descending=True)
            score_masked_row = score_masked_row[:new_added_cnts]
            global_new_added_score_list.append(score_masked_row)

        global_new_added_scores = torch.cat(global_new_added_score_list, dim=0)
        if global_new_added_scores.size()[0] == 0:
            # an empty tensor, at target sparsity is 0 situation
            return pre_masks
        threshold, _ = torch.kthvalue(global_new_added_scores, kth_masked_position, dim=0)
        for key in scores.keys():
            new_added_mask = new_added_masks[key]
            score = scores[key]
            new_added_filter = 1 - new_added_mask
            score_masked = (score * new_added_filter).abs()
            zero = torch.tensor([0.]).to(score.device)
            one = torch.tensor([1.]).to(score.device)
            progressive_mask = (new_added_mask + torch.where(score_masked <= threshold, zero, one)) * pre_masks[key]
            progressive_masks[key] = progressive_mask
        return progressive_masks


@register_pattern('N:M')
class PatternNInM(BasePattern):
    """Pruning Pattern.
    
    A Pattern class derived from Pattern. In this pattern, N out of every M continuous weights will be pruned.
    For more info of this pattern, please refer to :
    https://github.com/intel/neural-compressor/blob/master/docs/sparsity.md
    
    Args:
        config: A config dict object that contains the pattern information.
        
    Attributes:
        N: The number of elements to be pruned in a weight sequence.
        M: The size of the weight sequence.
    """

    def __init__(self, config, modules):
        """Initialize the basic pruning unit of N:M pattern."""
        super(PatternNInM, self).__init__(config, modules)
        pattern = self.pattern.split('_')[-1]
        self.N = int(pattern.split(':')[0])
        self.M = int(pattern.split(':')[1])  ##m is bigger
        self.check_layer_validity(self.modules, (self.N, self.M))

    def check_layer_validity(self, datas: dict, block_size: tuple):
        """Check if a layer is valid for this block_size.
        
        Args:
            datas: A dict object containing the weights for all layers.
            block_size: A tuple representing the size of the pattern block.
        """
        self.invalid_layers = []
        for key in datas.keys():
            data = datas[key].weight
            data = self._reshape_orig_to_2dims(data)
            shape = data.shape
            if shape[1] % block_size[1] != 0:
                self.invalid_layers.append(key)
                logger.warning(f"{key} shape {shape} cannot be divided by {self.pattern}")

    def get_reduced_masks_from_data(self, data, key):
        """Obtain the unpruned weights and reshape according to the block_size.
        
        Args:
            data: Input.
            key: The layer name.

        Returns:
            A tensor representing the unpruned weights.
        """
        data = self._reshape_orig_to_2dims(data)
        shape = data.shape
        M = self.M
        N = self.N
        new_shape = [shape[0], shape[1] // M, M]
        data = data.reshape(new_shape)
        nonzeros = torch.count_nonzero(data, dim=-1)
        reduced_mask = nonzeros > N
        return reduced_mask

    def get_least_ninm_mask_from_data(self, score):
        """Generate the least N scores in M.
        
        Args:
            score: the pruning scores of weights.

        Returns:
            A dict with the identical size as pre_masks and its 0/1 values are updated. 
                1 means unpruned and 0 means pruned.
        """
        current_score = score
        M = self.M
        N = self.N
        current_score = self._reshape_orig_to_2dims(current_score)
        shape = current_score.shape
        new_shape = [shape[0], shape[1] // M, M]
        current_score_new = current_score.reshape(new_shape)

        threshold, _ = torch.kthvalue(current_score_new, N, dim=2)
        threshold = threshold.unsqueeze(-1)

        threshold = threshold.expand(shape[0], shape[1] // M, M)
        threshold = threshold.reshape((shape[0], shape[1]))

        one = torch.tensor([1.]).to(current_score.device)
        zero = torch.tensor([0.]).to(current_score.device)
        mask = torch.where(current_score <= threshold, zero, one)
        return mask

    def get_sparsity_ratio(self, pre_masks, return_dict=False):
        """Please note that the zero cnt and total cnt are all block_wise for supporting channel-wise pruning.

        The return sparsity ratio is elementwised.
        
        Args:
            pre_masks: Dict{"layer_name": Tensor} that stores the masks generated after the last pruning step.
            return_dict: A bool determining whether to return more information like zero_cnt and total_cnt.
            
        Returns:
            An elementwise sparisty ratio.
        """
        zero_cnt = 0
        total_cnt = 0
        for key in pre_masks.keys():
            if key in self.invalid_layers:
                # total_cnt += pre_masks[key].numel() // self.M
                continue
            reduced_mask = self.get_reduced_masks_from_data(pre_masks[key], key)
            zero_cnt += int((torch.sum(reduced_mask == 0)).data.item())
            total_cnt += int(reduced_mask.numel())
        sparsity_ratio = float(zero_cnt) / total_cnt * self.N / self.M

        if return_dict:
            return {"sparsity_ratio": sparsity_ratio, "zero_cnt": zero_cnt,
                    "total_cnt": total_cnt}
        else:
            return sparsity_ratio

    def _reshape_orig_to_2dims(self, data):
        """Process layers that are not two-dimensional(e.g conv layer).

        Args:
            data: Input.
            
        Returns:
            Reshaped data.
        """
        if len(data.shape) == 4:  ##TODO need to verify whether it's ok for transposed conv
            data = data.permute(0, 2, 3, 1)  ##cout,k,k,cin
            data = data.reshape(data.shape[0], -1)
        return data

    def _reshape_2dims_to_orig(self, data, orig_shape):
        """Recover layers that are not two-dimensional(e.g conv layer).

        Args:
            data: Input.
            
        Returns:
            Reshaped data.
        """
        if len(orig_shape) == 4:
            data = data.reshape(orig_shape[0], orig_shape[2], orig_shape[3], orig_shape[1])
            data = data.permute(0, 3, 1, 2)
        return data

    def reshape_orig_to_pattern(self, data, key):  # pragma: no cover
        """Reshape the data based on the pruning pattern.
        
        Args:
            data: Input.
            key: layer name.
        
        Returns:
            Reshaped data.
        """
        data = self._reshape_orig_to_2dims(data)
        shape = data.shape
        new_shape = [shape[0], shape[1] // self.M, self.M]
        data = data.reshape(new_shape)
        return data

    def reshape_reduced_to_orig(self, data, key, orig_shape):  # pragma: no cover
        """Reshape the reduced data to its original shape.
        
        Args:
            data: Input.
            key: The layer name.
            orig_shape: The original shape of the layer.
            
        Returns:
            Data of its original shape.
        """
        data = data.repeat_interleave(self.M, dim=-1)
        return self._reshape_2dims_to_orig(data, orig_shape)

    def reduce_scores(self, scores):
        """Calculate the pruning scores after reducing the data and obtain the least N scores in M.
        
        Args:
            scores: Pruning scores of weights.

        Returns:
            Updated pruning scores and the least N scores in M.
        """
        ##to get the least N scores in M
        M = self.M
        N = self.N
        least_ninm_masks = {}
        new_scores = {}
        for key in scores.keys():
            if key in self.invalid_layers:  # pragma: no cover
                continue
            if self.keep_mask_layers.get(key, False):  # pragma: no cover
                continue
            current_score = scores[key]
            mask = self.get_least_ninm_mask_from_data(current_score)
            current_score_new = self._reshape_orig_to_2dims(current_score)
            shape = current_score_new.shape
            current_score_new = current_score_new.reshape((shape[0], shape[1]))
            ##to get the sum of N scores in each block with M
            current_score_new = current_score_new * (1.0 - mask)
            current_score_new = current_score_new.reshape(shape[0], shape[1] // M, M)
            score_sum = self.reduce_tensor(current_score_new, dim=-1)
            least_ninm_masks[key] = mask
            new_scores[key] = score_sum
        return new_scores, least_ninm_masks

    def get_ele_mask_per_threshold(self, score, threshold, block_size, least_ninm_mask):
        """Get the elementwise mask per threshold.

        Args:
            score: A tensor that stores the pruning scores of weights.
            threshold: A float used to determine whether to prune a weight.
            block_size: A list of two integers representing the height and width of the block.
            least_m_in_m_masks: A tensor representing the least N scores in M.
            
        Returns:
            mask: The elementwise pruning mask.
        """
        zero = torch.tensor([0.]).to(score.device)
        one = torch.tensor([1.]).to(score.device)
        mask = torch.where(score <= threshold, zero, one)
        mask = mask.repeat_interleave(block_size[1], dim=-1)
        ## both zero will be zero
        mask = (mask + least_ninm_mask)
        mask = torch.where(mask <= 0, zero, one)
        return mask

    def get_masks_global(self, scores, cur_target_sparsity_ratio, pre_masks,
                         keep_exact_sparsity_ratio=True):
        """Generate masks for layers.
        
        Gather all layer's scores together and calculate a common threshold.
        This threshold will be applied for all layers.
        
        Args:
            scores: A dict{"layer_name": Tensor} that stores the pruning scores of weights.
            target_sparsity_ratio: A float representing the model's final sparsity.
            pre_masks: A dict{"layer_name": Tensor} representing the masks generated after the last pruning step.
            max_sparsity_ratio_per_op: A float representing the maximum sparsity that one layer can reach.
            
        Returns:
            A dict with the identical size as pre_masks and its 0/1 values are updated. 
                1 means unpruned and 0 means pruned.
        """
        masks = pre_masks

        block_sparsity_ratio = cur_target_sparsity_ratio * self.M / self.N
        k_blockwise = self.update_residual_cnt(pre_masks, block_sparsity_ratio)
        if k_blockwise <= 0:
            return masks
        new_scores, least_ninm_masks = self.reduce_scores(scores)
        global_scores = torch.cat([torch.flatten(v) for v in new_scores.values()])  ##block_wise
        residual_k = k_blockwise
        not_exceed_layers = [key for key in new_scores.keys()]

        while True:
            threshold, _ = torch.kthvalue(global_scores, residual_k)
            for key in not_exceed_layers:
                score = new_scores[key]
                mask = self.get_ele_mask_per_threshold(score, threshold, (self.N, self.M), least_ninm_masks[key])
                info = self.get_sparsity_ratio({key: mask}, return_dict=True)
                zero_cnt = info["zero_cnt"]
                total_cnt = info["total_cnt"]
                current_sparsity_ratio = float(zero_cnt) / total_cnt
                key_new_sparsity = SparsityInfo(zero_cnt, total_cnt, current_sparsity_ratio)
                need_adjust, adjust_ratio = self.adjust_ratio(masks, key, key_new_sparsity,
                                                              self.max_sparsity_ratio_per_op * self.M / self.N,
                                                              self.min_sparsity_ratio_per_op * self.M / self.N,
                                                              self.target_sparsity_ratio * self.M / self.N)

                if need_adjust:  # pragma: no cover
                    self.keep_mask_layers[key] = True
                    masks[key] = self.get_single_mask_per_target_ratio(new_scores[key], adjust_ratio)
                    masks[key] = masks[key].repeat_interleave(self.M, dim=-1)
                    ## both zero will be zero
                    masks[key] = (masks[key] + least_ninm_masks[key])
                    zero = torch.tensor([0.]).to(score.device)
                    one = torch.tensor([1.]).to(score.device)
                    masks[key] = torch.where(masks[key] <= 0, zero, one)
                    if keep_exact_sparsity_ratio:
                        zero_cnt = self.get_sparsity_ratio({key: masks[key]}, return_dict=True)["zero_cnt"]
                        residual_k -= zero_cnt
                else:
                    masks[key] = mask
            if not keep_exact_sparsity_ratio:  # pragma: no cover
                break
            new_not_exceed_layers = [key for key in new_scores.keys() if not self.keep_mask_layers.get(key, False)]
            if not_exceed_layers == new_not_exceed_layers or len(new_not_exceed_layers) == 0:  # pragma: no cover
                break
            not_exceed_layers = new_not_exceed_layers
            global_scores = torch.cat([torch.flatten(new_scores[key]) for key in not_exceed_layers])

        for key in masks.keys():
            if key in self.invalid_layers:  # pragma: no cover
                continue
            if len(scores[key].shape) == 4:  ## need to permute
                mask = masks[key]
                orig_shape = scores[key].shape
                mask = self._reshape_2dims_to_orig(mask, orig_shape)
                masks[key] = mask
            layer_ratio = torch.sum(masks[key] == 0.0).data.item() / masks[key].numel()
            logger.info(f'layer {key} sparsity_ratio is {layer_ratio}')
        return masks

    def get_pattern_lock_masks(self, modules):
        """Obtain masks from original weight map, by masking where weights' are zero.
        
        Args:
            modules: A dict{"layer_name": Tensor} that stores weights.
            
        Returns:
            A dict with the identical size as modules, containing pattern lock masks.
        """
        pattern_lock_masks = {}
        for key in modules.keys():
            weight = modules[key].weight
            orig_shape = weight.shape
            if key in self.invalid_layers:
                mask = torch.ones(orig_shape, device=weight.device)
                pattern_lock_masks[key] = mask
                continue
            mask = self.get_least_ninm_mask_from_data(weight)
            mask = self._reshape_2dims_to_orig(mask, orig_shape)
            pattern_lock_masks[key] = mask
        return pattern_lock_masks



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The CastTo Pattern."""

from .pattern import Pattern, pattern_registry
from .. import graph_utils as util
import copy


@pattern_registry(pattern_type='CastTo')
class CastTo(Pattern):
    """The CastTo pattern.

    Fuse the original sub-graph into the custom acceleration 'CastTo' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'SoftmaxCast': [
                # softmax + cast
                {
                    'patterns': {
                        'in': [[(0, 'Softmax'), (1, 'Cast')]],
                        'out':[[(0, 'Softmax')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1]
                },
            ],
            'GreaterCastReduceSum': [
                # greater + cast + reduce_sum
                {
                    'patterns': {
                        'in': [[(0, 'Greater'), (1, 'Cast'), (2, 'ReduceSum')]],
                        'out':[[(0, 'Greater'), (1, 'ReduceSum')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2]
                },
            ],
            'RangeCastLess': [
                # range + cast + less
                {
                    'patterns': {
                        'in': [[(0, 'Cast'), (1, 'Range'), (2, 'Less')]],
                        'out':[[(0, 'Range'), (1, 'Less')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            2: [1]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [1, 0, 2]
                },
            ],
            'LessCastEqual': [
                # less + cast + equal
                {
                    'patterns': {
                        'in': [[(0, 'Less'), (1, 'Cast'), (2, 'Equal')]],
                        'out':[[(0, 'Less'), (1, 'Equal')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 2
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }], [[0, 1], 2]],
                        1: [[{
                            2: [1]
                        }], [[1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            0: [0]
                        }], [[0], 1]],
                        1: [[{
                            2: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1, 2]
                },
            ],
            'BatchMatMulCast': [
                # batch_matmul + cast
                {
                    'patterns': {
                        'in': [[(0, 'BatchMatMul'), (1, 'Cast')]],
                        'out':[[(0, 'BatchMatMul')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }], [[0, 1], 2]]
                    },
                    'output_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1]
                },
            ],
            'MatMulWithBiasCast': [
                # matmul_with_bias + cast
                {
                    'patterns': {
                        'in': [[(0, 'MatMulWithBias'), (1, 'Cast')]],
                        'out':[[(0, 'MatMulWithBias')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1
                    },
                    'input_tensors': {
                        0: [[{
                            0: [0]
                        }, {
                            0: [1]
                        }, {
                            0: [2]
                        }], [[0, 1, 2], 3]]
                    },
                    'output_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]]
                    },
                    'returns': [0, 1]
                },
            ],
        }

        def _set_attr(attr_list, cast_dtype, node_names, model):
            assert len(attr_list) == len(node_names)
            for i in range(len(node_names)):
                node_idx = model.get_node_id(node_names[i])
                node_attr = attr_list[i]
                node_type = model.nodes[node_idx].op_type
                if i == 0:
                    for k, v in cast_dtype.items():
                        if k[:3] == "dst":
                            if v == "fp32":
                                node_attr['dtype'] = v
                            else:
                                if node_type in ["Greater", "Less"]:
                                    if v in ["int32", "int64"]:
                                        node_attr['dtype'] = "s8"
                        else:
                            if node_type in ["Range"]:
                                if v in ["int64"]:
                                    node_attr[k] = "int32"
                            else:
                                node_attr[k] = v
                model.nodes[node_idx].attr = copy.deepcopy(node_attr)

        _fusion_order = [('SoftmaxCast', 2),
                         ('BatchMatMulCast', 2),
                         ('MatMulWithBiasCast', 2),
                         ('GreaterCastReduceSum', 2),
                         ('RangeCastLess', 2),
                         ('LessCastEqual', 2),
                        ]
        for pattern_key, keep_attr_start in _fusion_order:
            pattern_dict = pattern_mapping_config[pattern_key][0]
            model, new_node_names, ret_old_nodes = \
                util.pattern_mapping(pattern_key, pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    node = ret_old_nodes[j][0]
                    node_attr = copy.deepcopy(node.attr)
                    tensor_info = {}
                    cast_dtype = {}
                    for idx, t in enumerate(node.input_tensors):
                        tensor_info[t.name] = 'src' + str(idx) + '_dtype'
                    tensor_info[node.output_tensors[0].name] = 'dst'
                    for n in ret_old_nodes[j][1:keep_attr_start]:
                        if n.output_tensors[0].name in tensor_info:
                            cast_dtype[tensor_info[n.output_tensors[0].name]] = n.attr['DstT']
                        else:
                            cast_dtype[tensor_info[n.input_tensors[0].name]] = n.attr['DstT']
                    attr_list = [node_attr]
                    attr_list.extend([n.attr for n in ret_old_nodes[j][keep_attr_start:]])
                    _set_attr(attr_list, cast_dtype, new_node_names[j], model)

        return model



from intel_extension_for_transformers.optimization.utils.utility_tf import get_filepath
import numpy as np
import os
import shutil
import tensorflow as tf 
import unittest
from datasets import load_dataset, load_metric
from intel_extension_for_transformers.optimization import (
    metrics,
    PrunerConfig,
    PruningConfig,
    TFOptimization
)
from transformers import (
    AutoTokenizer,
    DefaultDataCollator,
    HfArgumentParser,
    TFAutoModelForSequenceClassification,
    TFTrainingArguments,
)

os.environ["WANDB_DISABLED"] = "true"


class TestTFPruning(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.model = TFAutoModelForSequenceClassification.from_pretrained(
            'bhadresh-savani/distilbert-base-uncased-sentiment-sst2'
        )
        raw_datasets = load_dataset("glue", "sst2")["validation"]
        tokenizer = AutoTokenizer.from_pretrained("bhadresh-savani/distilbert-base-uncased-sentiment-sst2")
        non_label_column_names = [name for name in raw_datasets.column_names if name != "label"]
        def preprocess_function(examples):
            # Tokenize the texts
            args = (
                (examples["sentence"],)
            )
            result = tokenizer(*args, padding=True, max_length=64, truncation=True)

            return result
        raw_datasets = raw_datasets.map(preprocess_function, batched=True, load_from_cache_file=False)
        data_collator = DefaultDataCollator(return_tensors="tf")
        dataset = raw_datasets.select(range(10))
        self.dummy_dataset = dataset.to_tf_dataset(
            columns=[col for col in dataset.column_names if col not in 
                     set(non_label_column_names + ["label"])],
            shuffle=False,
            batch_size=2,
            collate_fn=data_collator,
            drop_remainder=False,
            # `label_cols` is needed for user-defined losses, such as in this example
            # datasets v2.3.x need "labels", not "label"
            label_cols=["labels"] if "label" in dataset.column_names else None,
        )
        parser = HfArgumentParser(TFTrainingArguments)
        self.args = parser.parse_args_into_dataclasses(args=["--output_dir", "./quantized_model",
                                                        "--per_device_eval_batch_size", "2"])[0]
        optimizer = tf.keras.optimizers.Adam(
            learning_rate=self.args.learning_rate,
            beta_1=self.args.adam_beta1,
            beta_2=self.args.adam_beta2,
            epsilon=self.args.adam_epsilon,
            clipnorm=self.args.max_grad_norm,
        )
        loss_fn = tf.keras.losses.SparseCategoricalCrossentropy(
            from_logits=True, reduction=tf.keras.losses.Reduction.SUM
        )
        metrics = ["accuracy"]
        self.model.compile(optimizer=optimizer, loss=loss_fn, metrics=metrics)

    @classmethod
    def tearDownClass(self):
        shutil.rmtree('./tmp', ignore_errors=True)
        shutil.rmtree('./quantized_model', ignore_errors=True)

    def test_tf_model_quant(self):
        # check whether it is possible to set distributed environment
        # only for coverage currently
        from intel_extension_for_transformers.optimization.utils.utility_tf import distributed_init
        distributed_init(["localhost:12345","localhost:23456"], "worker", 0)
        self.assertTrue(os.environ['TF_CONFIG'] != None)
        del os.environ['TF_CONFIG']
        # check whether filepath can be set correctly if using distributed environment
        # only for coverage currently
        from intel_extension_for_transformers.optimization.utils.utility_tf import get_filepath
        self.assertTrue(type(get_filepath("dummy", "worker", 0)) == str)
        self.assertTrue(type(get_filepath("dummy", "worker", 1)) == str)
        self.assertTrue(get_filepath("dummy", "worker", 0) != get_filepath("dummy", "worker", 1))

        metric = load_metric("glue", "sst2")
        def compute_metrics(preds, label_ids):
            preds = preds["logits"]
            preds = np.argmax(preds, axis=1)
            result = metric.compute(predictions=preds, references=label_ids)
            if len(result) > 1:
                result["combined_score"] = np.mean(list(result.values())).item()
            return result
        self.optimizer = TFOptimization(
            model=self.model,
            args=self.args,
            train_dataset=self.dummy_dataset,
            eval_dataset=self.dummy_dataset,
            compute_metrics=compute_metrics,
        )
        tune_metric = metrics.Metric(
            name="accuracy", greater_is_better=True, is_relative=True, criterion=0.01,
        )
        prune_type = 'BasicMagnitude'
        target_sparsity_ratio = 0.1
        pruner_config = PrunerConfig(prune_type=prune_type, target_sparsity_ratio=target_sparsity_ratio)
        pruning_conf = PruningConfig(
            epochs=int(1), pruner_config=pruner_config, metrics=tune_metric
        )
        p_model = self.optimizer.prune(pruning_config=pruning_conf)
        loaded_model = tf.saved_model.load(self.args.output_dir)
        p_model = self.optimizer.prune(pruning_config=pruning_conf,
                                train_dataset=self.dummy_dataset,
                                eval_dataset=self.dummy_dataset,)

        def eval_func(model):
            return 1

        def train_func(model):
            return model
        
        self.optimizer.prune(pruning_config=pruning_conf,
                             train_func=train_func, 
                             eval_func=eval_func)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Pruning: specify the supported pruning mode."""

from packaging import version
from enum import Enum
from neural_compressor.conf.config import Pruner as INCPruner
from typing import Dict, List
from neural_compressor import __version__ as nc_version


class PruningMode(Enum):
    """Currently support three pruning modes."""
    BASICMAGNITUDE = "basic_magnitude"
    PATTERNLOCK = "pattern_lock"
    GROUPLASSO = "group_lasso"


SUPPORTED_PRUNING_MODE = set([approach.name for approach in PruningMode])


class PrunerConfig(INCPruner):
    """Pruner configuration."""
    def __init__(self, epoch_range: List=[0, 4], initial_sparsity_ratio: float=0.0,
                 target_sparsity_ratio: float=0.97, update_frequency: int=1,
                 prune_type: str='BasicMagnitude', method: str='per_tensor',
                 names: List=[], parameters: Dict=None):
        """Init the pruner config.

        Args:
            epoch_range: A list with length of 2. The first element is the start epoch and the second element
                is the end epoch. Pruning will be done from the start epoch to the end epoch.
            initial_sparsity_ratio: Initial sparsity goal
            target_sparsity_ratio: Target sparsity goal
            update_frequency: How many epochs to update once
            prune_type: "BasicMagnitude", "PatternLock", or "GroupLasso"
            method: TODO (Remove this parameter)
            names: A list of layer names that need to be pruned
            parameters: A dictionary of extra parameters
        """
        if epoch_range is not None:
            assert len(epoch_range) == 2, "Please set the epoch_range as [start_epoch, end_epoch]"
            self.start_epoch = epoch_range[0]
            self.end_epoch = epoch_range[1]
        else:
            self.start_epoch = None
            self.end_epoch = None
        self.update_frequency = update_frequency
        self.target_sparsity = target_sparsity_ratio
        self.initial_sparsity = initial_sparsity_ratio
        self.update_frequency = update_frequency
        assert prune_type.upper() in SUPPORTED_PRUNING_MODE, \
            "prune_type only support {}!".format(
                [mode.lower() for mode in SUPPORTED_PRUNING_MODE]
            )
        self.prune_type = PruningMode[prune_type.upper()].value
        self.method = method
        self.names= names
        self.parameters = parameters






#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The neural engine empty op file."""

from .op import Operator, operator_registry, list2str, parseTorchListConstruct
from .tensor import Tensor


# x + y element-wise, supports broadcasting
@operator_registry(operator_type='AddV2')
class AddV2(Operator):
    """Register the AddV2 operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='MultiHeadAttenion')
class MultiHeadAttenion(Operator):

    def __init__(self):
        super().__init__()


# x + y element-wise, supports broadcasting
@operator_registry(operator_type='Add')
class Add(Operator):
    """Register the Add operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='BinaryAdd')
class BinaryAdd(Operator):
    """Register the BinaryAdd operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='ConstantOfShape')
class ConstantOfShape(Operator):
    """Register the ConstantOfShape operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='DequantizeLinear')
class DequantizeLinear(Operator):
    """Register the DequantizeLinear operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Dequantize')
class Dequantize(Operator):
    """Register the DequantizeLinear operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Computes the Gauss error function of x element-wise.
@operator_registry(operator_type='Erf')
class Erf(Operator):
    """Register the Erf operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Expand')
class Expand(Operator):
    """Register the Expand operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# This operation creates a tensor of shape dims and fills it with value.
# tf.fill(dims, value, name=None)
@operator_registry(operator_type='Fill')
class Fill(Operator):
    """Register the Fill operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='FlatMapDataset')
class FlatMapDataset(Operator):
    """Register the FlatMapDataset operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Identity')
class Identity(Operator):
    """Register the Identity operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op MatMul + BiasAdd
# The inputs are two-dimensional matrices and 1-D const bias
@operator_registry(operator_type='InnerProduct')
@operator_registry(operator_type='InnerProductGraph')
class InnerProduct(Operator):
    """Register the InnerProduct operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# store input_tensors for engine
@operator_registry(operator_type='Input')
class Input(Operator):
    """Register the Input operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='MakeIterator')
class MakeIterator(Operator):
    """Register the MakeIterator operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op MatMulWithBias + Add/AddV2
# The inputs are two-dimensional matrices, 1-D const bias and one tensor from Add op
@operator_registry(operator_type='MatMulWithBiasAdd')
class MatMulWithBiasAdd(Operator):
    """Register the MatMulWithBiasAdd operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op MatMulWithBias + Gelu
# The inputs are two-dimensional matrices and 1-D const bias
@operator_registry(operator_type='MatMulWithBiasGelu')
class MatMulWithBiasGelu(Operator):
    """Register the MatMulWithBiasGelu operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op MatMulWithBias + Tanh
# The inputs are two-dimensional matrices and 1-D const bias
@operator_registry(operator_type='MatMulWithBiasTanh')
class MatMulWithBiasTanh(Operator):
    """Register the MatMulWithBiasTanh operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op MatMul + BiasAdd
# The inputs are two-dimensional matrices and 1-D const bias
@operator_registry(operator_type='MatMulWithBias')
class MatMulWithBias(Operator):
    """Register the MatMulWithBias operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# store the output_tensors for engine
@operator_registry(operator_type='Output')
class Output(Operator):
    """Register the Output operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='QLinearMatMul')
class QLinearMatMul(Operator):
    """Register the QLinearMatMul operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='QLinearAdd')
class QLinearAdd(Operator):
    """Register the QLinearAdd operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='QLinearMul')
class QLinearMul(Operator):
    """Register the QLinearMul operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Returns x / y element-wise for real types.
# If x and y are reals, this will return the floating-point division.
# RealDiv supports broadcasting
@operator_registry(operator_type='RealDiv')
class RealDiv(Operator):
    """Register the RealDiv operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# tf.math.rsqrt(x, name=None)
# Computes reciprocal of square root of x element-wise.
@operator_registry(operator_type='Rsqrt')
class Rsqrt(Operator):
    """Register the Rsqrt operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()

# Computes element-wise square root of the input tensor.
@operator_registry(operator_type='Sqrt')
class Sqrt(Operator):
    """Register the Sqrt operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Computes square of x element-wise.
@operator_registry(operator_type='Square')
class Square(Operator):
    """Register the Square operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# tf.math.squared_difference(x, y, name=None)
# Returns conj(x - y)(x - y) element-wise, supports broadcasting
@operator_registry(operator_type='SquaredDifference')
class SquaredDifference(Operator):
    """Register the SquaredDifference operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='StopGradient')
class StopGradient(Operator):
    """Register the StopGradient operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Given an input tensor, this function computes hyperbolic tangent of every element in the tensor.
@operator_registry(operator_type='Tanh')
class Tanh(Operator):
    """Register the Tanh operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='TensorSliceDataset')
class TensorSliceDataset(Operator):
    """Register the TensorSliceDataset operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


# Fused_op Reshape, Transpose and BatchMatMul / BatchMatMulV2
# This pattern has several ops combinations, so the input_tensors and output_tensors may various
@operator_registry(operator_type='TransposeBatchMatMul')
class TransposeBatchMatMul(Operator):
    """Register the TransposeBatchMatMul operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Where')
class Where(Operator):
    """Register the Where operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Range')
class Range(Operator):
    """Register the Range operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Relu')
class Relu(Operator):
    """Register the Relu operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='MatMulWithBiasRelu')
class MatMulWithBiasRelu(Operator):
    """Register the MatMulWithBiasRelu operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Matmul')
class Matmul(Operator):
    """Register the Matmul operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='CumSum')
class CumSum(Operator):
    """Register the CumSum operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Onehot')
class Onehot(Operator):
    """Register the Onehot operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='TokenTypeIds')
class TokenTypeIds(Operator):
    """Register the TokenTypeIds operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='PositionIds')
class PositionIds(Operator):
    """Register the PositionIds operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Loop')
class Loop(Operator):
    """Register the Loop operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Sigmoid')
class Sigmoid(Operator):
    """Register the Sigmoid operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='MatMulWithBiasSigmoid')
class MatMulWithBiasSigmoid(Operator):
    """Register the MatMulWithBiasSigmoid operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='EmbeddingBag')
class EmbeddingBag(Operator):
    """Register the EmbeddingBag operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Flatten')
class Flatten(Operator):
    """Register the Flatten operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='MergedEmbeddingbag')
class MergedEmbeddingbag(Operator):
    """Register the MergedEmbeddingbag operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Convolution')
class Convolution(Operator):
    """Register the Convolution operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='ExpandIndices')
class ExpandIndices(Operator):
    """Register the ExpandIndices operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='Tile')
class Tile(Operator):
    """Register the Tile operator."""

    def __init__(self):
        """The init function of this operator."""
        super().__init__()


@operator_registry(operator_type='OpAny')
class OpAny(Operator):
    """Register the OpAny operator."""

    def __init__(self):
        """The init function of this operator."""


@operator_registry(operator_type='SequenceLength')
class SequenceLength(Operator):
    """Register the SequenceLength operator."""
    def __init__(self):
        super().__init__()

@operator_registry(operator_type='Arange')
class Arange(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Reciprocal')
class Reciprocal(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Einsum')
class Einsum(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Repeat')
class Repeat(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Stack')
class Stack(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Constant')
class Constant(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='ListConstruct')
class ListConstruct(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='LatRange')
class LatRange(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Masked_fill')
class Masked_fill(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Floor_divide')
class Floor_divide(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Max')
class Max(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='ListUnpack')
class ListUnpack(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()

@operator_registry(operator_type='Silu')
class Silu(Operator):
    """Register the Tile operator."""
    def __init__(self):
        """The init function of this operator."""
        super().__init__()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""The TextEncoder_KVReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import OrderedDict
from .. import graph_utils as util
from .. import logger


@pattern_registry(pattern_type='TextEncoder_KVReshape')
class TextEncoder_KVReshape(Pattern):
    """The TextEncoder_KVReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'TextEncoder_KVReshape' graph.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'TextEncoder_KVReshape': [
                # for text encoder k_proj and v_proj (reshape 0 & 1)
                {
                    'patterns': {
                        'in': [[(0, 'Unsqueeze'), (2, 'Concat'), (3, 'Reshape'), (4, 'Transpose')],
                               [(), (1, 'MatMulWithBias'), (3, 'Reshape')]],
                        'out': [[(0, 'MatMulWithBias'), (1, 'Reshape'), (2, 'Transpose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 1,
                        1: 3,
                        2: 4
                    },
                    'input_tensors': {
                        0: [[{
                            1: [0]
                        }, {
                            1: [1]
                        }, {
                            1: [2]
                        }], [[0, 1, 2], 3]],
                        1: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        2: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0: [[{
                            1: [0]
                        }], [[0], 1]],
                        1: [[{
                            3: [0]
                        }], [[0], 1]],
                        2: [[{
                            4: [0]
                        }], [[0], 1]],
                    },
                    'returns': [2, 1, 4]
                },
            ]
        }

        def _set_attr(head_num, head_size, node_names, model):
            attr = OrderedDict()
            attr['dst_shape'] = '-1,-1,' + str(head_num) + ',' + str(head_size)
            attr['dims'] = '0'

            reshape_node_idx = model.get_node_id(node_names[1])
            assert model.nodes[reshape_node_idx].op_type == 'Reshape'
            model.nodes[reshape_node_idx].attr = attr

        def _keep_attr(ret_old_nodes, node_names, model):
            output_node_names_idx = 0
            ret_old_names_idx = 1
            MatMulWithBias_node_idx = model.get_node_id(node_names[output_node_names_idx])
            assert ret_old_nodes[ret_old_names_idx].op_type == model.nodes[MatMulWithBias_node_idx].op_type
            model.nodes[MatMulWithBias_node_idx].attr = ret_old_nodes[ret_old_names_idx].attr

            output_node_names_idx = 2
            ret_old_names_idx = 2
            transpose_node_idx = model.get_node_id(node_names[output_node_names_idx])
            assert ret_old_nodes[ret_old_names_idx].op_type == model.nodes[transpose_node_idx].op_type
            model.nodes[transpose_node_idx].attr = ret_old_nodes[ret_old_names_idx].attr

        for i in range(len(pattern_mapping_config['TextEncoder_KVReshape'])):
            pattern_dict = pattern_mapping_config['TextEncoder_KVReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("TextEncoder_KVReshape", pattern_dict,
                                                                        model)

            logger.info('TextEncoder_KVReshape mathched...')
            logger.debug('TextEncoder_KVReshape = {}'.format(new_node_names))
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    pack_node = ret_old_nodes[j][0]
                    head_size = int(pack_node.input_tensors[-1].data)  # 32
                    head_num = int(pack_node.input_tensors[-2].data)  # 12
                    _set_attr(head_num, head_size, new_node_names[j], model)
                    _keep_attr(ret_old_nodes[j], new_node_names[j], model)
                return model

        return model



"""scheduler module."""
# !/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

SCHEDULERS = {}


def register_scheduler(name):
    """Class decorator used to register a Scheduler subclass to the registry.
    
    Decorator function used before a Scheduler subclass.
    Make sure that the Scheduler class decorated by this function can be registered in SCHEDULERS.
    
    Args:
        cls (class): The class of register.
        name: A string that defines the scheduler type.
        
    Returns:
        cls: The class of register.
    """
    def register(scheduler):
        SCHEDULERS[name] = scheduler
        return scheduler

    return register


def get_scheduler(config):
    """Get registered scheduler class.
    
    Get a scheduler object from SCHEDULERS.
    
    Args:
        config: A config dict object that contains the scheduler information.
        
    Returns:
        A Scheduler object.
    """
    name = "iterative"
    if config.start_step == config.end_step:  # pragma: no cover
        name = "oneshot"
    return SCHEDULERS[name](config)


class PruningScheduler:
    """Pruning Scheduler.

    The class which defines a sparsity changing process during pruning.
    Mainly contains two types:
        1. iterative scheduler. Prune the model from dense to target sparsity gradually.
        2. one-shot scheduler. Prune the model in a single step and reach the target sparsity.
        
    Args:
        config: A config dict object that contains the scheduler information.
        
    Attributes:
        config: A config dict object that contains the scheduler information.
    """
    
    def __init__(self, config):
        """Initialize."""
        self.config = config

    def update_sparsity_ratio(self, target_ratio, current_prune_step, total_prune_steps, masks, init_ratio=0.0):
        """To be implemented in subclasses."""
        raise NotImplementedError


@register_scheduler('oneshot')
class OneshotScheduler(PruningScheduler):  # pragma: no cover
    """Pruning Scheduler.
    
    A Scheduler class derived from Scheduler.
    Prune the model to target sparsity once.
    
    Args:
        config: A config dict object that contains the scheduler information.
        
    Attributes:
        Inherit from parent class Scheduler.
    """

    def __init__(self, config):
        """Initialize."""
        super(OneshotScheduler, self).__init__(config)

    def update_sparsity_ratio(self, target_ratio, current_prune_step, total_prune_steps, masks, init_ratio=0.0):
        """Update sparsity ratio.
        
        Args:
            target_ratio: A float representing the sparsity ratio after pruning.
            current_prune_step: An integer representing the current pruning step.
            total_prune_steps: An integer representing the total number of steps of the pruning process.
            masks: A dict {"module_name": Tensor} that stores the masks for modules' weights.
            init_ratio: A float representing the sparsity ratio before pruning. 

        Return:
            A float representing the sparsity ratio that the model will reach after the next pruning step.
        """
        return target_ratio


@register_scheduler('iterative')
class IterativeScheduler(PruningScheduler):
    """Pruning Scheduler.
    
    A Scheduler class derived from Scheduler.
    Prune the model from dense to target sparsity in several steps.
    
    Args:
        config: A config dict object that contains the scheduler information.

    Attributes:
        Inherit from parent class Scheduler.
    """

    def __init__(self, config):
        """Initialize."""
        super(IterativeScheduler, self).__init__(config)

    def update_sparsity_ratio(self, target_ratio, current_prune_step, total_prune_steps, masks,
                              init_sparsity_ratio=0.0):
        """Obtain new target sparsity ratio according to the step.

        Args:
            target_ratio: A float. The target sparsity ratio.
            current_prune_step: A integer. The current pruning step.
            total_prune_steps: A integer. The total steps included in the pruning progress.
            masks: A dict{"module_name": Tensor}. The masks for modules' weights.
            init_sparsity_ratio:
        
        Returns:
            A float representing the target sparsity ratio the model will reach after the next pruning step.
        """
        aggressive_ratio = target_ratio
        aggressive_ratio = min(self.config.max_sparsity_ratio_per_op,
                               aggressive_ratio)  ##legacy issue

        decay_type = self.config.sparsity_decay_type
        if decay_type == "cos":
            current_target_sparsity = (aggressive_ratio - init_sparsity_ratio) * (
                    1.0 - math.cos(float(current_prune_step) / total_prune_steps * (math.pi / 2))) + init_sparsity_ratio
        elif decay_type == "exp":
            target_dense_change_ratio = ((1.0 - aggressive_ratio) / (1.0 - init_sparsity_ratio)) ** (
                    1 / total_prune_steps)
            current_target_sparsity = 1.0 - (
                    1.0 - init_sparsity_ratio) * target_dense_change_ratio ** current_prune_step

        elif decay_type == "linear":  # pragma: no cover
            current_target_sparsity = (aggressive_ratio - init_sparsity_ratio) * float(
                current_prune_step) / total_prune_steps + init_sparsity_ratio

        elif decay_type == "cube":  # pragma: no cover
            current_target_sparsity = (aggressive_ratio - init_sparsity_ratio) * (
                    (float(current_prune_step) / total_prune_steps) ** 3) + init_sparsity_ratio
        else:  # pragma: no cover
            assert False, "{} is not supported".format(decay_type)

        current_target_sparsity = min(target_ratio, current_target_sparsity)
        return current_target_sparsity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The DecoderAttnReshape Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
import copy
from .. import graph_utils as util


@pattern_registry(pattern_type='DecoderAttnReshape')
class DecoderAttnReshape(Pattern):
    """The DecoderAttnReshape pattern.

    Fuse the original sub-graph into the custom acceleration 'DecoderAttnReshape' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'DecoderAttnReshape': [
                {
                    'patterns': {
                        'in': [[(1, 'Shape'), (2, 'Gather'), (3, 'Unsqueeze'), (7, 'Concat'),
                                (8,'Reshape'), (9, 'Gather'), (10, 'Transpose')],
                               [(), (4, 'Shape'), (5, 'Gather'), (6, 'Unsqueeze'), (7, 'Concat')],
                               [(), (0, 'Unsqueeze'), (7, 'Concat')]],
                        'out': [[(0, 'Reshape'), (1, 'Gather'), (2, 'Transpose')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 8,
                        1: 9,
                        2: 10
                    },
                    'input_tensors': {
                        0: [[{
                            8: [0]
                        }, {
                            'input_data': [0]
                        }], [[0, 1], 2]],
                        1: [[{
                            9: [1]
                        }], [[0], 2]],
                        2: [[], [[], 1]],
                    },
                    'output_tensors': {
                        0 :[[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            10: [0]
                        }], [[0], 1]]
                    },
                    'returns': [7, 9, 10]
                },
            ]
        }

        def _set_attr(head_num, node_names, old_nodes, model, reshape_pos=0, keep_from=1):
            attr = OrderedDict()
            attr['dst_shape'] = '-1,' + str(head_num) + ',-1,-1'
            attr['dims'] = '1,0'

            reshape_node_idx = model.get_node_id(node_names[reshape_pos])
            model.nodes[reshape_node_idx].attr = attr
            keep_idx = keep_from
            while keep_idx < len(old_nodes):
                node_idx = model.get_node_id(node_names[keep_idx])
                model.nodes[node_idx].attr = copy.deepcopy(old_nodes[keep_idx].attr)
                keep_idx += 1

        for i in range(len(pattern_mapping_config['DecoderAttnReshape'])):
            pattern_dict = pattern_mapping_config['DecoderAttnReshape'][i]
            model, new_node_names, ret_old_nodes = util.pattern_mapping("ContentAttnReshape",
                                                                        pattern_dict, model)
            if len(new_node_names) != 0:
                for j in range(len(new_node_names)):
                    pack_node = ret_old_nodes[j][0]
                    head_num = int(pack_node.input_tensors[1].data)
                    _set_attr(head_num, new_node_names[j], ret_old_nodes[j], model)
                return model

        return model



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The AttentionBlock_WeightReshapeTo4D Pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util
from .subgraph_matcher import EXECUTOR_TYPE
import numpy as np
import copy

@pattern_registry(pattern_type='AttentionBlock_WeightReshapeTo4D')
class AttentionBlock_WeightReshapeTo4D(Pattern):
    """The AttentionBlock_WeightReshapeTo4D pattern.

    Fuse the original sub-graph into the custom acceleration 'AttentionBlock_WeightReshapeTo4D' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """

    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'AttentionBlock_WeightReshapeTo4D': [
                {
                    'patterns': {
                        'in': [[(0, 'Reshape'), (1, 'Mul'), (2, 'Add'), (3, 'Sigmoid')]],
                    },
                },
            ]
        }

        pattern = pattern_mapping_config['AttentionBlock_WeightReshapeTo4D'][0]['patterns']['in']
        patterns_nodes_name = util.search_pattern(pattern, model)
        print('AttentionBlock_WeightReshapeTo4D = ', patterns_nodes_name)

        if len(patterns_nodes_name) != 0:
            for j in range(len(patterns_nodes_name)):
                add_node = model.get_node_by_name(patterns_nodes_name[j][2])
                add_node.input_tensors[1].shape = [1, 512, 1, 1]

        return model




#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Quantization: specify the supported quantization mode."""

from enum import Enum
from transformers.utils.versions import require_version

try:
    require_version("neural_compressor>=1.9.0")
except:
    require_version("neural_compressor_full>=1.9.0", "To fix: pip install neural_compressor")


class QuantizationMode(Enum):
    """Currently support three quantization modes."""
    POSTTRAININGSTATIC = "post_training_static_quant"
    POSTTRAININGDYNAMIC = "post_training_dynamic_quant"
    QUANTIZATIONAWARETRAINING = "quant_aware_training"


SUPPORTED_QUANT_MODE = set([approach.name for approach in QuantizationMode])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The PositionEmbeddingsV1 pattern."""

from .pattern import Pattern, pattern_registry
from collections import namedtuple, OrderedDict
from .. import graph_utils as util


@pattern_registry(pattern_type='PositionEmbeddingsV1')
class PositionEmbeddingsV1(Pattern):
    """The PositionEmbeddingsV1 pattern.

    Fuse the original sub-graph into the custom acceleration 'PositionEmbeddingsV1' graph.
    The search strategy is based on the following pattern mapping configs for different models.
    """
    def __call__(self, model):
        """The __call__ function of this pattern class."""
        pattern_mapping_config = {
            'PositionEmbeddingsV1': [
                # roberta_base
                {
                    'patterns': {
                        'in': [[(0, 'Equal'), (1, 'Not'), (2, 'Cast'), (3, 'CumSum'), 
                                (4, 'Add'), (5, 'Mul'), (6, 'Cast'), (7, 'Add'), (8, 'Gather')]],
                        'out': [[(0, 'PositionIds'), (1, 'Reshape'),
                                 (2, 'Gather'), (3, 'Reshape'), (4, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/Add_40',
                        1: 'position_embeddings/reshape',
                        2: 8,
                        3: 'position_embeddings/after/reshape',
                        4: 'position_embeddings/add_reshape',
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            8: [0]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        4: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            8: [0]
                        }], [[0], 1]]
                    },
                    'returns': [8]
                },
                # roberta_base int8
                {
                    'patterns': {
                        'in': [[(0, 'Equal'), (1, 'Not'), (2, 'Cast'), (3, 'CumSum'), 
                                 (4, 'Mul'), (5, 'Cast'), (6, 'Add'), (7, 'Gather')]],
                        'out': [[(0, 'PositionIds'), (1, 'Reshape'),
                                 (2, 'Gather'), (3, 'Reshape'), (4, 'Reshape')]]
                    },
                    'search_mode': 'op_type',
                    'node_names': {
                        0: 'position_embeddings/Add_40',
                        1: 'position_embeddings/reshape',
                        2: 7,
                        3: 'position_embeddings/after/reshape',
                        4: 'position_embeddings/add_reshape',
                    },
                    'input_tensors': {
                        0: [[{
                            'input_data': [0]
                        }], [[0], 1]],
                        1: [[], [[], 1]],
                        2: [[{
                            7: [0]
                        }], [[1], 2]],
                        3: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                        4: [[{
                            'input_data': [0]
                        }], [[1], 2]],
                    },
                    'output_tensors': {
                        0: [[], [[], 1]],
                        1: [[], [[], 1]],
                        2: [[], [[], 1]],
                        3: [[], [[], 1]],
                        4: [[{
                            7: [0]
                        }], [[0], 1]]
                    },
                    'returns': [7]
                },
            ]
        }

        def _set_attr(hidden_size, axis, batch_dims, node_names, model):
                attr0 = OrderedDict()
                attr0['mode'] = "roberta"
                attr1 = OrderedDict()
                attr1['dst_shape'] = -1
                attr2 = OrderedDict()
                attr2['axis'] = axis
                attr2['batch_dims'] = batch_dims
                attr3 = OrderedDict()
                attr3['dst_shape'] = '-1,-1,' + str(hidden_size)
                attr3['dims'] = '0,1'
                attr4 = OrderedDict()
                attr4['dst_shape'] = '-1,-1,' + str(hidden_size)
                attr4['dims'] = '0,1'
                attr4['mul'] = '1,2'
                
                position_node_idx = model.get_node_id(node_names[0])
                model.nodes[position_node_idx].attr = attr0

                reshape_0_node_idx = model.get_node_id(node_names[1])
                model.nodes[reshape_0_node_idx].attr = attr1

                gather_node_idx = model.get_node_id(node_names[2])
                model.nodes[gather_node_idx].attr = attr2

                reshape_1_node_idx = model.get_node_id(node_names[3])
                model.nodes[reshape_1_node_idx].attr = attr3

                reshape_2_node_idx = model.get_node_id(node_names[4])
                model.nodes[reshape_2_node_idx].attr = attr4

        # bert_roberta_base
        for i in range(len(pattern_mapping_config['PositionEmbeddingsV1'])):
            pattern_dict = pattern_mapping_config['PositionEmbeddingsV1'][i]
            model, new_node_names, ret_old_nodes = \
                util.pattern_mapping('PositionEmbeddingsV1', pattern_dict, model)
            if len(new_node_names) != 0:
                for i in range(len(new_node_names)):
                    gatherv2_node = ret_old_nodes[i][0]
                    hidden_size = int(gatherv2_node.input_tensors[0].shape[-1])
                    axis = gatherv2_node.attr['axis']
                    batch_dims = gatherv2_node.attr['batch_dims']
                    _set_attr(hidden_size, axis, batch_dims, new_node_names[i], model)

        return model
